library control_keyboard;

uses
 {$IFDEF WIN32}
  ShareMem,
 {$ENDIF}
  SysUtils,
  Classes,
  o3control,
  dglOpenGL,
  o3actors,
  o3vectors,
  o3game,
  o3interfaces,
  Math;

{$LIBPREFIX ''}

type TKeyboardControl = class;
     TControlListener = class(TO3InputListener)
      owner:TKeyboardControl;
      procedure keyDown(key:TO3Key); override;
      procedure keyUp(key:TO3Key); override;
     end;
     TKeyboardControl = class(To3Control)
      KeyboardDeltaX,KeyboardDeltaY:integer;
      KeyboardSensivity:single;
      keysPressed:array[TO3Key] of integer;   // 0 = released, 1 = pressed,
      buttonsPressed:array[0..10] of integer; //    2 = fresh pressed
      Listener:TControlListener;
      procedure KeyDown(key:TO3Key);
      procedure KeyUp(key:TO3Key);
      function Steer(ac:pointer;t:integer):TVector3d; override;
      constructor Create;
      destructor Destroy; override;
     end;
var Ozone:TOzoneMainClass;

procedure TControlListener.keyDown;
begin
 owner.keyDown(key);
end;

procedure TControlListener.keyUp;
begin
 owner.keyUp(key);
end;


function MakeKeyboardControl:TO3Control;
begin
 Result:=TKeyboardControl.Create;
end;

procedure RegisterLib(entry:pointer;o3_state:pointer);
begin
 ozone:=o3_state;
 PControlEntry(entry).MakeControl:=MakeKeyboardControl;
end;

 procedure TKeyboardControl.KeyDown;
 begin
  keysPressed[key]:=2;
 end;

 procedure TKeyboardControl.KeyUp;
 begin
  keysPressed[key]:=0;
 end;

 constructor TKeyboardControl.Create;
 begin
  inherited Create;
  listener:=TControlListener.Create;
  listener.owner:=Self;
  ozone.InputSys.addListener(listener);
  KeyboardDeltaX:=0;
  KeyboardDeltaY:=0;
  KeyboardSensivity:=ozone.O3Ini.ReadFloat('control_keyboard','sensivity',1);
 end;


 function TKeyboardControl.Steer;
 var s:integer;
     actor:TActor;
     c:boolean;
 const MinSens = 1;
       MaxSens = 30;
 begin
  actor:=TActor(ac);
  if actor=nil then Exit;
  if (keysPressed[O3K_2]>0) then begin
   if (keysPressed[O3K_A]>1) then begin
    actor.Look:=(actor.Look+1)mod LCount;
    keysPressed[O3K_A]:=1;
   end;
   if (keysPressed[O3K_D]>1) then begin
    actor.Look:=(actor.Look+LCount-1)mod LCount;
    keysPressed[O3K_D]:=1;
   end;
  end;
  KeyboardDeltaX:=0;
  KeyboardDeltaY:=0;
  if keysPressed[O3K_W]>0 then
   KeyboardDeltaY:=-t;
  if keysPressed[O3K_S]>0 then
   KeyboardDeltaY:=t;
  if keysPressed[O3K_A]>0 then
   KeyboardDeltaX:=-t;
  if keysPressed[O3K_D]>0 then
   KeyboardDeltaX:=t;
  actor.canTakeUpItems:=true;
  if keysPressed[O3K_1]>1 then begin
   keysPressed[O3K_1]:=1;
   actor.LookFar:=not actor.LookFar;
  end;
  if keysPressed[O3K_E]>0 then begin
   if keysPressed[O3K_E]>1 then
    actor.UseFirstItem;
   keysPressed[O3K_E]:=1;
   actor.canTakeUpItems:=false;
  end;

  if keysPressed[O3K_R]>0 then begin
   if keysPressed[O3K_R]>1 then
    actor.RotateItems;
   keysPressed[O3K_R]:=1;
   actor.canTakeUpItems:=false;
  end;
  c:=false;
  if (keysPressed[O3K_3] > 1) then begin KeyboardSensivity:=KeyboardSensivity+1; c:=true; keysPressed[O3K_3]:=1; end;
  if (keysPressed[O3K_4] > 1) then begin KeyboardSensivity:=KeyboardSensivity-1; c:=true; keysPressed[O3K_4]:=1; end;
  if KeyboardSensivity<MinSens then KeyboardSensivity:=MinSens;
  if KeyboardSensivity>MaxSens then KeyboardSensivity:=MaxSens;
  if c then
   actor.Inventary.StartStaticText('Sensivity: '+Copy(FloatToStr(KeyboardSensivity),1,3),1000);
  if L[actor.Look].SwapXY then begin
   s:=KeyboardDeltaX;
   KeyboardDeltaX:=KeyboardDeltaY;
   KeyboardDeltaY:=s;
  end;
   Result := getVector(
                       sign(KeyboardDeltaX)*power(abs(KeyboardDeltaX)*0.4,1.2)*
                       L[actor.Look].XS*KeyboardSensivity,
                       0,
                       sign(KeyboardDeltaY)*power(abs(KeyboardDeltaY)*0.4,1.2)*
                       L[actor.Look].YS*KeyboardSensivity);
  KeyboardDeltaX:=0;
  KeyboardDeltaY:=0;
 end;


 destructor TKeyboardControl.Destroy;
 begin
//  ozone.InputSys.deleteKeyboardMoveListener(KeyboardMove);
//  ozone.InputSys.deleteKeyDownListener(keyDown);
//  ozone.InputSys.deleteKeyUpListener(keyUp);
//  ozone.InputSys.deleteKeyboardDownListener(KeyboardDown);
//  ozone.InputSys.deleteKeyboardUpListener(KeyboardUp);
  ozone.InputSys.deleteListener(listener);
  listener.Free;
  ozone.O3Ini.WriteFloat('control_keyboard','sensivity',KeyboardSensivity);
  inherited Destroy;
 end;

exports RegisterLib;
begin
end.
