library actor_marbles;

uses
 {$IFDEF WIN32}
  ShareMem,
 {$ENDIF}
  SysUtils,
  Classes,
  o3actors,
  o3utils,
  o3world,
  ac3d,
  odeimport,
  o3control,
  o3stones,
  o3item,
  md2,
  o3game,
  o3physics,
  Math,
  o3m;

type
  TDir = (dn,de,ds,dw);
  PDir = ^TDir;
  TDestructAnim = class(TGameObject)
   md:TQuakeModel;
   ac:TActor;
   frame:single;
   procedure callBack(d:integer); override;
  end;
  TBallType = (firstplayercount,nocount,BlackBall,WhiteBall,SmallBall);
  TMarble = class(TActor)
   MarbleType:TBallType; // The type to count lives
   TypeID:TBallType; // The REAL type, mostly (but not always) equal to MarbleType
   isSpawning:boolean;
   spawnFrame:Single;
   spawnFPS:integer;
   procedure CreateBodyAndGeom(c:string;var g:TGeomList;var b:PdxBody); override;
   procedure ProcessOvers; override;
   procedure processMessage(m:string;p:array of pointer); override;
   procedure OnStoneOver(stone:TStone;level:Pointer);
   procedure Render(timeDelta:integer); override;
   destructor Destroy; override;
   constructor Create(name:string;sx,sy,sz:Single;control:TO3Control;Own:Pointer;i:string;coll:string); override;
   procedure PickUpItem(Item:TItem;level:Pointer); override;
   procedure DropItem(level:Pointer;x,y,z:integer); override;
   function tryRebirth:boolean; override;
   procedure Spawn(s:boolean); override;
  end;
  TBlackBall = class(TMarble)
   procedure Render(t:integer); override;
   constructor Create(name:string;sx,sy,sz:Single;control:TO3Control;Own:Pointer;i:string;coll:string); override;
  end;
  TWhiteBall = class(TMarble)
   procedure Render(t:integer); override;
   constructor Create(name:string;sx,sy,sz:Single;control:TO3Control;Own:Pointer;i:string;coll:string); override;
  end;
  TSmallBall = class(TMarble)
   procedure Render(t:integer); override;
   constructor Create(name:string;sx,sy,sz:Single;control:TO3Control;Own:Pointer;i:string;coll:string); override;
  end;



  var p:string;
      ozone:TOzoneMainClass;
      BlackBallGfx:TO3mFile;
      ShatterGfx:TQuakeModel;
      WhiteBallGfx:TO3mFile;
      SmallBallGfx:TO3mFile;
      Prot,ProtEnd:TO3mFile;
//      destructAnim:TQuakeModel;
      spawnAnim:TQuakeModel;
      Lives:array[firstplayercount..SmallBall] of Integer;
{$LIBPREFIX 'lib'}

 procedure TMarble.processMessage;
 var t:PBoolean;
     d:PDir;
 begin
  inherited;
  if m='marblecall' then begin
   PInteger(p[0])^:=0;
   case TypeID of
    BlackBall: PInteger(p[0])^:=1;
    WhiteBall: PInteger(p[0])^:=2;
    SmallBall: PInteger(p[0])^:=3;
   end;
  end;
  if m='laserhit' then begin
   d:=p[0];
   if (TypeID=SmallBall)and
      (
      ((((x-trunc(x))<0.3)or((x-trunc(x))>0.7))and((d^=dn)or(d^=ds)))or
      (((y-trunc(y))<0.3)or((y-trunc(y))>0.7))or
      (((z-trunc(z))<0.3)or((z-trunc(z))>0.7)and((d^=de)or(d^=dw)))
      ) then begin
    t:=p[1];
    t^:=true;   // do not cut the beam
    EXIT;
   end;
   if ((y-trunc(y))<0.2)or((y-trunc(y))>0.8) then begin
    t:=p[1];
    t^:=true;   // do not cut the beam
    EXIT;
   end;
   processMessage('kill',[]);
  end;
 end;

 procedure TMarble.Spawn;
 begin
  isSpawning:=true;
  spawnFrame:=0;
  if s then spawnFPS:=10
  else spawnFPS:=15;
  PODEBodyData(dBodyGetData(body)).fix[0]:=x;
  PODEBodyData(dBodyGetData(body)).fix[1]:=y;
  PODEBodyData(dBodyGetData(body)).fix[2]:=z;
 end;

 procedure TDestructAnim.callBack;
 begin
 if Frame<(md.NumFrames)-1 then begin
  md.Frame:=frame;
  md.animationStep(d/1000,md.NumFrames,false);
  frame:=md.Frame;
  md.Material:=md2mat(TO3mFile(ac.gfx).FPalette[0]);
  md.UseMaterial:=true;
  if TMarble(ac).TypeID=smallball then begin
   md.Scale[0]:=0.35;
   md.Scale[1]:=0.35;
   md.Scale[2]:=0.35;
//   write('s');
  end
  else begin
   md.Scale[0]:=1;
   md.Scale[1]:=1;
   md.Scale[2]:=1;
  end;
  ozone.gfxSys.DrawModel(md,cx,cy,cz,quake,ac.gfxinfo);
 end
 else begin
  md.Frame:=md.NumFrames-1;
  Frame:=Frame+((d/1000)*md.NumFrames);
  if Frame>20 then begin
   tLevel(owner).CallBackBuffer.FList.Remove(self);
   self.Free;
   exit;
  end
  else begin
   md.Material:=md2mat(TO3mFile(ac.gfx).FPalette[0]);
   md.Material.Alpha:=md.Material.Alpha-((Frame-md.NumFrames)/(20-md.NumFrames));
   md.UseMaterial:=true;
   ozone.gfxSys.DrawModel(md,cx,cy,cz,quake,ac.gfxinfo);
  end;
  end;
 end;

 procedure TMarble.PickUpItem;
 var Inv:pointer;
 begin
  Inv:=Inventary;
  if MarbleType=firstplayercount then
   Inventary:=TLevel(owner).theActorToFollow.Inventary;
  inherited PickUpItem(item,level);
  Inventary:=Inv;
 end;

 procedure TMarble.DropItem;
 begin
  if not (MarbleType=firstplayercount) then
   inherited DropItem(level,x,y,z);
 end;

 constructor TBlackBall.Create;
 begin
//  Self.ClassName+Self.
  ozone.O3Log.LogStatus('blackball creation, stage 1','TBlackBall.Create');
  MarbleType:=BlackBall;
  TypeID:=BlackBall;
  physics.selfFric:=0;
  physics.mass:=0.3;
  physics.steerPower:=0.4;
//  write(' s2');
  ozone.O3Log.LogStatus('blackball creation, stage 2','TBlackBall.Create');
  inherited Create('ac-blackball',sx,sy,sz,control,Own,i,'s0.35');
  gfxType:=o3utils.ozone;
  gfx:=BlackBallGfx;
  ozone.O3Log.LogStatus('blackball creation, finished','TBlackBall.Create');
  shadow:=ozone.ShadowLib['sh-blackball'];
 end;

 constructor TWhiteBall.Create;
 begin
  ozone.O3Log.LogStatus('whiteball creation, stage 1','TWhiteBall.Create');
  MarbleType:=WhiteBall;
  TypeID:=WhiteBall;
  physics.selfFric:=0;
  physics.mass:=0.3;
  physics.steerPower:=0.4;
  ozone.O3Log.LogStatus('whiteball creation, stage 2','TWhiteBall.Create');
  inherited Create('ac-whiteball',sx,sy,sz,control,Own,i,'s0.35');
  gfx:=WhiteBallGfx;
  gfxType:=o3utils.ozone;
  ozone.O3Log.LogStatus('whiteball creation finished','TWhiteBall.Create');
  // same shadow as blackball
   shadow:=ozone.ShadowLib['sh-blackball'];
 end;

 constructor TSmallBall.Create;
 begin
  ozone.O3Log.LogStatus('smallball creation, stage 1','TSmallBall.Create');
  MarbleType:=SmallBall;
  TypeID:=SmallBall;
  physics.selfFric:=0;
  physics.mass:=0.1;
  physics.steerPower:=0.4;
  ozone.O3Log.LogStatus('smallball creation, stage 2','TSmallBall.Create');
  inherited Create('ac-smallball',sx,sy,sz,control,Own,i,'s0.125');
  gfx:=SmallBallGfx;
  gfxType:=o3utils.ozone;
  ozone.O3Log.LogStatus('smallball creation finished','TSmallBall.Create');
  shadow:=ozone.ShadowLib['sh-smallball'];
 end;

 constructor TMarble.Create;
 var x:integer;
     c:string;
 begin
  x:=1;
//  write('token get');
  if i<>'' then begin
   c:=GetToken(i,x,',');
//   write('token got');
   while c<>'' do begin
//    writeln('loop, c=',c);
    if c='nolifecount' then begin
     MarbleType:=nocount;
//     writeln('nolifecount');
    end;
    if c='firstplayercount' then begin
     MarbleType:=firstplayercount;
//     writeln('1st count');
    end;
    if (c<>'')and(c[1]='C') then begin
     if Copy(c,2,255)='blackball' then
      MarbleType:=BlackBall
     else if Copy(c,2,255)='whiteball' then
      MarbleType:=WhiteBall
     else if Copy(c,2,255)='smallball' then
      MarbleType:=SmallBall;
    end;
    inc(x);
    c:=GetToken(i,x,',');
   end;
  end;
  inherited Create(name,sx,sy,sz,control,Own,i,coll);
  if not (marbletype in [nocount,firstplayercount]) then
   if Lives[MarbleType]>0 then
    for x:=1 to Lives[MarbleType] do
     PickUpItem(TLevel(own).MakeItem('it-extralife',own),own);
 end;

 destructor TMarble.Destroy;
 var x:integer;
 begin
  if not (marbletype in [nocount,firstplayercount]) then begin
   if OutOfGame or (not(TLevel(owner).state in [levWon,levRestart])) then
    Lives[MarbleType]:=2
   else begin
    Lives[MarbleType]:=0;
    if Inventary<>nil then
     if Inventary.Count>0 then
      for x:=0 to Inventary.Count-1 do
       if Inventary[x]<>nil then
        if TItem(Inventary[x]).Info['extralife']='yes' then
         inc(Lives[MarbleType]);
   end;
  end;
  inherited;
 end;

 procedure TBlackBall.Render;
 begin
    inherited;
    if visible and (not isSpawning) then ozone.gfxSys.DrawModel(BlackBallGfx,cx,cy,cz,o3utils.ozone,gfxinfo);
    if physics.godFor>2000 then ozone.gfxSys.DrawModel(Prot,cx,cy,cz,o3utils.ozone,gfxinfo)
     else if physics.godFor>1 then ozone.gfxSys.DrawModel(ProtEnd,cx,cy,cz,o3utils.ozone,gfxinfo);
 end;

 procedure TSmallBall.Render;
 begin
    inherited;
    if visible and (not isSpawning) then ozone.gfxSys.DrawModel(SmallBallGfx,cx,cy,cz,o3utils.ozone,gfxinfo);
    if physics.godFor>2000 then ozone.gfxSys.DrawModel(Prot,cx,cy,cz,o3utils.ozone,gfxinfo)
     else if physics.godFor>1 then ozone.gfxSys.DrawModel(ProtEnd,cx,cy,cz,o3utils.ozone,gfxinfo);
 end;

 procedure TWhiteBall.Render;
 begin
    inherited;
    if visible and (not isSpawning) then ozone.gfxSys.DrawModel(WhiteBallGfx,cx,cy,cz,o3utils.ozone,gfxinfo);
    if physics.godFor>2000 then ozone.gfxSys.DrawModel(Prot,cx,cy,cz,o3utils.ozone,gfxinfo)
     else if physics.godFor>1 then ozone.gfxSys.DrawModel(ProtEnd,cx,cy,cz,o3utils.ozone,gfxinfo);
 end;

 procedure TMarble.CreateBodyAndGeom;
 var m:TdMass;
     a:PdxGeom;
//     p:pptrInfo;
     t:PODEBodyData;
 begin
 //  ozone.O3Log.LogStatus('callback, info='+c,'TMarble.CreateBodyAndGeom');
   c:=Copy(c,2,255);
   dMassSetSphere(m,1,strToFloat(c));
   dMassAdjust(m,physics.mass);
   dBodySetMass(body,@m);
   dBodySetPosition(body,x,y,z);
   ozone.O3Log.LogStatus('creating sphere','TMarble.CreateBodyAndGeom');
   a:=dCreateSphere(TLevel(owner).Physics.theSpace,strToFloat(c));
//   a:=dCreateBox(TLevel(owner).Physics.theSpace,1,1,1);
   ozone.O3Log.LogStatus('setting geom-body','TMarble.CreateBodyAndGeom');
   dGeomSetBody(a,body);
 //  new(p);
//   p^.info:='actor';
//   p^.P:=self;
   dGeomSetData(a,self);
   dGeomSetPosition(a,x,y,z);
   geoms.add(a);
   t:=NewODEBodyData;
   dBodySetData(body,t);
//   ozone.O3Log.LogStatus('!ok!','TMarble.CreateBodyAndGeom');
 end;

// constructor TMarble.Create;
// begin
//    inherited Create(name,sx,sy,sz,control,Own,i);
// end;


 procedure InitLib(path:string;owner:Pointer);
 var t:TBallType;
 begin
  DecimalSeparator:='.';
  p:=path;
  ozone:=owner;
  BlackBallGfx:=TO3mFile.Create(ozone);
  BlackBallGfx.LoadFromFile(p+'ac-blackball.o3m');
  ShatterGfx:=TQuakeModel.Create(ozone);
  ShatterGfx.LoadModel(p+'marble-shatter.md2','material_marble.bmp');
  WhiteBallGfx:=TO3mFile.Create(ozone);
  WhiteBallGfx.LoadFromFile(p+'ac-whiteball.o3m');
  SmallBallGfx:=TO3mFile.Create(ozone);
  SmallBallGfx.LoadFromFile(p+'ac-smallball.o3m');
  Prot:=TO3mFile.Create(ozone);
  Prot.LoadFromFile(p+'ac-marble-prot.o3m');
  ProtEnd:=TO3mFile.Create(ozone);
  ProtEnd.LoadFromFile(p+'ac-marble-prot-end.o3m');
//  destructAnim:=TQuakeModel.Create(ozone);
//  destructAnim.LoadModel(p+'marble-break.md2','material_marble.bmp');
  spawnAnim:=TQuakeModel.Create(ozone);
  spawnAnim.LoadModel(p+'marble-spawn.md2','material_marble.bmp');
//  destructAnim.AutoCalcVertexNormals;
  for t:=Low(Lives) to High(Lives) do
   Lives[t]:=2;
  Lives[nocount]:=0;
 end;

 procedure QuitLib;
 begin
  BlackBallGfx.free;
  ShatterGfx.free;
  WhiteBallGfx.Free;
  SmallBallGfx.Free;
  Prot.Free;
  ProtEnd.Free;
//  destructAnim.Free;

 end;

 function MakeActor(s:string;sx,sy,sz:Single;control:TO3Control;Own:Pointer;i:string):TActor;
 begin
  Result:=nil;
//  write('ball:');
  if s='ac-blackball' then begin
//    write('bb');
    Result:=TBlackBall.Create(s,sx,sy,sz,control,own,i,'');
   end
  else if s='ac-whiteball' then begin
//    write('wb');
    Result:=TWhiteBall.Create(s,sx,sy,sz,control,own,i,'');
  end
  else if s='ac-smallball' then begin
//    write('sb');
    Result:=TSmallBall.Create(s,sx,sy,sz,control,own,i,'');
  end
 end;

 procedure TMarble.OnStoneOver;
 begin
  alive:=false;
 end;


 procedure TMarble.ProcessOvers;
 var sx,sy,sz:integer;
 begin
  if TLevel(Owner)=nil then Exit;
  sx:=floor(cx);
  sy:=floor(cy);
  sz:=floor(cz);
  if TLevel(Owner).getCell(sx,sy,sz)=nil then Exit;
  if TLevel(Owner).getCell(sx,sy,sz).FItem<>nil then
   TLevel(Owner).getCell(sx,sy,sz).FItem.OnActorOver(self,owner,sx,sy,sz);
  if TLevel(Owner).getCell(sx,sy,sz).FStone<>nil then
   if (Assigned(TLevel(Owner).getCell(sx,sy,sz).FStone.collisionCallback))and(TLevel(Owner).getCell(sx,sy,sz).FStone.collisionCallback(0,TLevel(Owner).getCell(sx,sy,sz).FStone,sx,sy,sz,self,owner,nil)) then
    OnStoneOver(TLevel(Owner).getCell(sx,sy,sz).FStone,owner);
 end;

 procedure TMarble.Render;
 var i:integer;
     t:Pdvector3;
 begin
  if isSpawning then begin
   spawnAnim.Frame:=spawnFrame;
   spawnAnim.animationStep(timeDelta/1000,spawnFPS);
   spawnFrame:=spawnAnim.Frame;
   if TypeID=SmallBall then begin
    spawnAnim.Scale[0]:=0.35;
    spawnAnim.Scale[1]:=0.35;
    spawnAnim.Scale[2]:=0.35;
   end
   else begin
    spawnAnim.Scale[0]:=1;
    spawnAnim.Scale[1]:=1;
    spawnAnim.Scale[2]:=1;
   end;
   spawnAnim.Material:=md2Mat(TO3mFile(gfx).FPalette[0]);
   ozone.gfxSys.DrawModel(spawnAnim,cx,cy,cz,quake,gfxinfo);
   if spawnFrame>8.9 then begin
     isSpawning:=false;
     spawnFrame:=0;
     PODEBodyData(dBodyGetData(body)).fix[0]:=odeb_inactive;
     PODEBodyData(dBodyGetData(body)).fix[1]:=odeb_inactive;
     PODEBodyData(dBodyGetData(body)).fix[2]:=odeb_inactive;
   end;

  end;
  if OutOfGame then begin
   visible:=false;
//   Draw(ProtEnd,cx,cy,cz);
  end;
  if TLevel(owner).state<>levWon then begin
   t:=dBodyGetLinearVel(body);
   if physics.godFor>0 then begin
    if physics.GodY<>-555 then
     y:=physics.GodY;
    dBodySetLinearVel(body,t[0],0,t[2]);
   end;
  end;
  if physics.godFor>0 then
   dec(physics.godFor,timeDelta);

  ProcessOvers;
  if y<-5 then
   ProcessMessage('kill',[]);
  if Inventary.count>0 then begin
   for i:=0 to Inventary.count-1 do begin
    TItem(Inventary[i]).processMessage('iTick',[@timeDelta,self]);
   end;
  end;
 end;

 function TMarble.tryRebirth;
 var inv:Pointer;
     i:integer;
     t:TDestructAnim;
 begin
  TOzoneMainClass(TLevel(owner).owner).O3Log.LogStatus('trying rebirth','TMarble.tryRebirth');
  if y>-5 then begin
   TOzoneMainClass(TLevel(owner).owner).O3Log.LogStatus('spawning anim at:'+floattostr(x)+','+floattostr(y)+','+floattostr(z),'TMarble.tryRebirth');
   t:=TDestructAnim.Create;
   t.cx:=x;
   if typeID=smallball then
    t.cy:=y-0.1
   else
    t.cy:=y-0.3;
   //t.cy:=y;
   t.cz:=z;
   t.frame:=2;
   t.ac:=self;
   t.md:=ShatterGfx;
   t.owner:=owner;
   TLevel(owner).CallBackBuffer.Add(t);
  end;
  inv:=Inventary;
  Result:=false;
  if MarbleType=firstplayercount then begin
   Inventary:=TLevel(owner).theActorToFollow.Inventary;
  end;
  if Inventary.count>0 then begin
   for i:=0 to Inventary.count-1 do
    if inventary[i]<>nil then
     if (TItem(inventary[i]).Info['extralife']='yes')and(not Result) then begin
      TItem(inventary[i]).free;
      inventary.Delete(i);
      Result:=true;
      alive:=true;
      dBodySetLinearVel(body,0,0,0);
      if flagx>0 then begin
       x:=flagx;
       y:=flagy;
       z:=flagz;
       break;
      end
      else begin
       x:=startx;
       y:=starty;
       z:=startz;
       break;
      end;
     end;
   end;
   if MarbleType=firstplayercount then begin
     Inventary:=inv;
     if not result then TLevel(owner).theActorToFollow.OutOfGame:=true;
     TLevel(owner).Notify1STdestroyed(result);
   end;
  TOzoneMainClass(TLevel(owner).owner).O3Log.LogStatus('trying rebirth okay','TMarble.tryRebirth');
 end;

 exports MakeActor,
         InitLib,
         QuitLib;

begin
end.
