library stone_puzzles;

uses
  {$IFDEF WIN32}ShareMem,{$ENDIF}
  SysUtils,
  Classes,
  Math,
  ac3d,
  o3m,
  o3utils,
  o3stones,
  o3world,
  o3game,
  o3actors,
  o3vectors,
  o3item,
  o3basicstones;

{$LIBPREFIX 'lib'}

type
    TPuzzleType = (ptRound,ptLine,ptT,ptEnd,ptCross,ptHole);
    TPuzzleConnection = (conN,conE,conS,conW,holed);
    TPuzzleConnections = set of TPuzzleConnection;
    TPuzzleCluster = class;
    TPuzzleStone = class(TStone)
     connections:TPuzzleConnections;
     tag:boolean;
     expCnt:single;
//     procedure explode;
     procedure Render(timeDelta:integer); override;
     procedure moveTo(tx,ty,tz:integer); override;
     function moveQuery(tx,ty,tz:integer):boolean; override;
     constructor create(s:string;own:pointer); reintroduce;
     procedure processMessage(m:string;data:array of pointer); override;
    end;
    TPuzzleCluster = class
     public
      FList:TList;
      isComplete:boolean;
      procedure addToCluster(stone:TPuzzleStone);
      procedure addAll(stone:TPuzzleStone);
      procedure removeFromCluster(stone:TPuzzleStone); overload;
      procedure removeFromCluster(cluster:TPuzzleCluster); overload;
      procedure transpose(dx,dy,dz:integer);
      constructor Create(stone:TPuzzleStone;all:boolean = false);
      destructor Destroy; override;
    end;

var PuzzleStoneGfx: array[TPuzzleType] of TO3mFile;

 function collFuncPuzzle(dir:integer;call:TStone;x,y,z:integer;actor:Pointer;level:Pointer;geom:pointer):boolean;
 var item:TItem;
     t,all:TPuzzleCluster;
     c:boolean;
 begin
  if TPuzzleStone(call).expCnt>0 then begin
   Result:=dir=0;
   exit;
  end;
  item:=TActor(actor).getFirstItem;
  if (item<>nil){and(item.info['wand']='yes')} then begin
//   TBlock_Break(call).isBreaking:=true;
//   ozone.soundLib.playSound('block_break',false);
    t:=TPuzzleCluster.Create(TPuzzleStone(Call));
    all:=TPuzzleCluster.Create(TPuzzleStone(Call),true);
//    writeln(t.flist.count,' ',all.flist.count);
    if (t.isComplete)and(t.flist.count=all.flist.count) then begin
     call.processMessage('explode',[]);
     Result:=true;
     t.Free;
     all.free;
    end else if (t.isComplete)and(t.flist.count<all.flist.count) then begin
     c:=true;
     all.removeFromCluster(t);
     t.free;
     while all.flist.count>0 do begin
      t:=TPuzzleCluster.Create(TPuzzleStone(all.Flist.First));
      c:=c and t.isComplete;
      all.removeFromCluster(t);
      t.free;
     end;
     if c then begin
      call.processMessage('explode',[]);
      Result:=true;
//      t.Free;
      all.free;
     end
     else begin
//      t.Free;
      all.free;
      if holed in TPuzzleStone(call).connections then
       Result:=collFuncHole(dir,call,x,y,z,actor,level,geom)
      else
       Result:=collFuncMove(dir,call,x,y,z,actor,level,geom);
     end;
    end else begin
     t.Free;
     all.free;
     if holed in TPuzzleStone(call).connections then
      Result:=collFuncHole(dir,call,x,y,z,actor,level,geom)
     else
      Result:=collFuncMove(dir,call,x,y,z,actor,level,geom);
    end;
  end
  else begin
   if holed in TPuzzleStone(call).connections then
    Result:=collFuncHole(dir,call,x,y,z,actor,level,geom)
   else
    Result:=collFuncMove(dir,call,x,y,z,actor,level,geom);
  end;
 end;

 function invertC(t:tPuzzleConnection):TPuzzleConnection;
 begin
  Result:=conN;
  case t of
   conN : Result:=conS;
   conE : Result:=conW;
   conS : Result:=conN;
   conW : Result:=conE;
  end;
 end;

 function tx(t:tPuzzleConnection):integer;
 begin
  Result:=0;
  case t of
   conN:Result:=0;
   conE:Result:=1;
   conS:Result:=0;
   conW:Result:=-1;
  end;
 end;

 function tz(t:tPuzzleConnection):integer;
 begin
  Result:=0;
  case t of
   conN:Result:=-1;
   conE:Result:=0;
   conS:Result:=1;
   conW:Result:=0;
  end;
 end;

 function simplifyType(t:TPuzzleConnections):TPuzzleType;
 begin
//  Result:=ptRound;
  if holed in t then begin Result:=ptHole; exit; end;
  if (t=[conN,conE])or(t=[conE,conS])or(t=[conS,conW])or(t=[conN,conW]) then Result:=ptRound
  else if (t=[conN,conE,conS])or(t=[conE,conS,conW])or(t=[conN,conS,conW])or(t=[conN,conE,conW]) then Result:=ptT
  else if (t=[conN])or(t=[conE])or(t=[conS])or(t=[conW]) then Result:=ptEnd
  else if (t=[conN,conS])or(t=[conE,conW]) then Result:=ptLine
  else result:=ptCross;
 end;

 function getRotation(t:TPuzzleConnections):integer;
 begin
//  Result:=0;
  if holed in t then begin Result:=0; exit; end;

  if (t=[conN,conE]) then Result:=0
  else if(t=[conE,conS]) then Result:=270
  else if(t=[conS,conW]) then Result:=180
  else if (t=[conN,conW]) then Result:=90

  else if (t=[conN]) then Result:=0
  else if(t=[conE]) then Result:=270
  else if(t=[conS]) then Result:=180
  else if (t=[conW]) then Result:=90

  else if (t=[conN,conE,conS]) then Result:=0
  else if(t=[conE,conS,conW]) then Result:=270
  else if(t=[conN,conS,conW]) then Result:=180
  else if (t=[conN,conE,conW]) then Result:=90

  else if (t=[conN,conS]) then Result:=0
  else if (t=[conE,conW]) then Result:=90
  else Result:=0;
 end;

 function puzzleConnectionsToStr(t:tPuzzleConnections):string;
 begin
  Result:='';
  if holed in t then begin Result:='hollow'; exit; end;
  if conN in t then Result:=Result+'n';
  if conE in t then Result:=Result+'e';
  if conS in t then Result:=Result+'s';
  if conW in t then Result:=Result+'w';
 end;

 function strToPuzzleConnections(t:string):TPuzzleConnections;
 begin
  result:=[];
  if pos('hollow',t)<>0 then begin result := [conN,conE,conS,conW,holed]; exit; end;
  if pos('n',t)<>0 then include(result,conN);
  if pos('e',t)<>0 then include(result,conE);
  if pos('s',t)<>0 then include(result,conS);
  if pos('w',t)<>0 then include(result,conW);
  if result=[] then result:=[conN,conE,conS,conW];
 end;

 procedure TPuzzleCluster.removeFromCluster(stone:TPuzzleStone);
 begin
  FList.Remove(stone);
  stone.info['cluster']:='';
  stone.info['visit']:='';
 end;

 procedure TPuzzleCluster.removeFromCluster(cluster:TPuzzleCluster);
 var x:integer;
 begin
  if cluster.flist.count<=0 then exit;
  for x:=0 to cluster.flist.count-1 do
   removeFromCluster(TPuzzleStone(cluster.flist[x]));
 end;

 procedure TPuzzleCluster.transpose;
 var p:TPuzzleStone;
     i:integer;
 begin
  if FList.Count=0 then exit;
  for i:=0 to Flist.Count-1 do begin
   p:=TPuzzleStone(Flist[i]);
   if not p.moveQuery(p.x+dx,p.y+dy,p.z+dz) then exit;
  end;
  for i:=0 to Flist.Count-1 do begin
   p:=TPuzzleStone(Flist[i]);
   TLevel(p.owner).getCell(p.x,p.y,p.z).FStone:=nil;
   p.x:=p.x+dx;
   p.y:=p.y+dy;
   p.z:=p.z+dz;
  end;
  for i:=0 to Flist.Count-1 do begin
   p:=TPuzzleStone(Flist[i]);
   TLevel(p.owner).getCell(p.x,p.y,p.z).FStone:=p;
   p.processMessage('move',[]);
  end;
  p:=TPuzzleStone(Flist.First);
  TOzoneMainClass(TLevel(p.owner).owner).soundLib.playSound('stone_move',p.cx,p.cy,p.cz);
 end;

 procedure TPuzzleCluster.addToCluster(stone:TPuzzleStone);
 var cell:TCell;
     other:TPuzzleStone;
     con:TPuzzleConnection;
//     adr:string;
 begin
  Flist.add(stone);
  stone.info['cluster']:='yes';
  for con:=conN to conW do begin
   if not(con in stone.connections) then begin
{    cell:=TLevel(stone.owner).getCell(stone.x+tx(con),stone.y,stone.z+tz(con));
    if cell=nil then continue;
    if cell.FStone=nil then continue;
    if cell.FStone.info['puzzle']='yes' then isComplete:=false;}
    continue;
   end;
   cell:=TLevel(stone.owner).getCell(stone.x+tx(con),stone.y,stone.z+tz(con));
   if cell=nil then begin isComplete:=false; continue; end;
   if cell.FStone=nil then begin isComplete:=false; continue; end;
   if cell.FStone.info['puzzle']<>'yes' then begin isComplete:=false; continue; end;
   other:=TPuzzleStone(cell.FStone);
   if not(invertC(con) in other.connections) then begin isComplete:=false; continue; end;
   if other.info['cluster']<>'yes' then begin
//    other.info['cluster']:=adr;
    addToCluster(other);
   end;
  end;
 end;

 procedure TPuzzleCluster.addAll(stone:TPuzzleStone);
 var cell:TCell;
     other:TPuzzleStone;
     con:TPuzzleConnection;
 begin
  Flist.add(stone);
  stone.info['visit']:='yes';
  for con:=conN to conW do begin
   cell:=TLevel(stone.owner).getCell(stone.x+tx(con),stone.y,stone.z+tz(con));
   if cell=nil then continue;
   if cell.FStone=nil then continue;
   if cell.FStone.info['puzzle']<>'yes' then continue;
   other:=TPuzzleStone(cell.FStone);
   if other.info['visit']<>'yes' then begin
    addAll(other);
   end;
  end;
 end;

 constructor TPuzzleCluster.Create;
 begin
  inherited Create();
  isComplete:=true;
  FList:=TList.Create;
  if all then addAll(stone)
  else addToCluster(stone);
 end;

 destructor TPuzzleCluster.Destroy;
 begin
  while Flist.Count>0 do begin
   removeFromCluster(TPuzzleStone(FList.last));
  end;
  FList.Clear;
  Flist.Free;
  inherited;
 end;


 procedure TPuzzleStone.processMessage;
 begin
  if m='explode' then
   if expCnt=-1 then expCnt:=0;
 end;

 constructor TPuzzleStone.create;
 var t:string;
 begin
  t:=GetToken(s,2,':');
  connections:=strToPuzzleConnections(t);
  if holed in connections then
   inherited Create(s,'hole',own)
  else
   inherited Create(s,'move',own);
  gfx:=PuzzleStoneGfx[simplifyType(connections)];
  gfxtype:=o3utils.ozone;
  info['puzzle']:='yes';
  info['connections']:=puzzleConnectionsToStr(connections);
  AntiForce:=3;
  tag:=false;
  collisionCallback:=collFuncPuzzle;
  expCnt:=-1;
  if holed in connections then
   shadow:=TOzoneMainClass(TLevel(own).owner).shadowLib['sh-puzzle-hole']
  else
   shadow:=TOzoneMainClass(TLevel(own).owner).shadowLib['sh-stone'];
 end;

 procedure TPuzzleStone.MoveTo;
 var t:TPuzzleCluster;
 begin
  t:=TPuzzleCluster.Create(self);
  t.transpose(tx-x,ty-y,tz-z);
  t.Free;
 end;

 function TPuzzleStone.MoveQuery;
 var t:TCell;
 begin
  if info['cluster']='' then begin
   Result:=true;
   Exit;
  end;
  Result:=inherited MoveQuery(tx,ty,tz);
  if not Result then begin
   if info['cluster']<>'yes' then exit;
   t:=TLevel(owner).getCell(tx,ty,tz);
   if t=nil then exit;
   if t.Fstone.info['puzzle']<>'yes' then exit;
   if t.Fstone.info['cluster']<>'yes' then exit;
   Result:=true;
  end;
 end;

 procedure TPuzzleStone.Render;
 var cell:TCell;
 begin
  if expCnt=0 then begin
   TLevel(owner).particleSystem.spawn(150,2,getVector(1,0,0),getVector(0.5,0.5,0.5),getVector(cx,cy,cz),1000,100,0,nil,nil);
   TLevel(owner).particleSystem.spawn(150,2,getVector(1,0.5,0),getVector(0.5,0.5,0.5),getVector(cx,cy,cz),1000,100,0,nil,nil);
   TOzoneMainClass(TLevel(owner).owner).soundLib.playSound('explosion',cx,cy,cz);
  end;
  if expCnt>=0.25 then begin
   cell:=TLevel(owner).getCell(x+1,y,z);
   if (cell<>nil)and(cell.FStone<>nil)and(cell.FStone.info['puzzle']='yes')and(TPuzzleStone(cell.Fstone).expCnt<0) then cell.FStone.processMessage('explode',[]);
   cell:=TLevel(owner).getCell(x-1,y,z);
   if (cell<>nil)and(cell.FStone<>nil)and(cell.FStone.info['puzzle']='yes')and(TPuzzleStone(cell.Fstone).expCnt<0) then cell.FStone.processMessage('explode',[]);
   cell:=TLevel(owner).getCell(x,y,z+1);
   if (cell<>nil)and(cell.FStone<>nil)and(cell.FStone.info['puzzle']='yes')and(TPuzzleStone(cell.Fstone).expCnt<0) then cell.FStone.processMessage('explode',[]);
   cell:=TLevel(owner).getCell(x,y,z-1);
   if (cell<>nil)and(cell.FStone<>nil)and(cell.FStone.info['puzzle']='yes')and(TPuzzleStone(cell.Fstone).expCnt<0) then cell.FStone.processMessage('explode',[]);
  end;
  if expCnt>=1 then begin
   if TLevel(owner).getCell(x,y,z)<>nil then
    TLevel(owner).getCell(x,y,z).Fstone:=nil;
   Free;
  end;
  if expCnt>-1 then expCnt:=expCnt+timeDelta/1000
  else begin
   TO3MFile(gfx).Fobject.kids[1].rotate[1]:=getRotation(connections);
   inherited Render(timeDelta);
   //TOzoneMainClass(TLevel(owner).owner).gfxSys.DrawModel(gfx,0,0,0,ozone);
  end;
 end;


 function MakeStone(s:string;o:Pointer):TStone;
 begin
  Result:=TPuzzleStone.create(s,o);
 end;

 procedure InitLib(path:string;owner:Pointer);
 begin
  PuzzleStoneGfx[ptRound]:=TO3mFile.Create(owner);
  PuzzleStoneGfx[ptRound].LoadFromFile(path+'st-puzzle-ne.o3m');
  PuzzleStoneGfx[ptT]:=TO3mFile.Create(owner);
  PuzzleStoneGfx[ptT].LoadFromFile(path+'st-puzzle-nes.o3m');
  PuzzleStoneGfx[ptEnd]:=TO3mFile.Create(owner);
  PuzzleStoneGfx[ptEnd].LoadFromFile(path+'st-puzzle-n.o3m');
  PuzzleStoneGfx[ptLine]:=TO3mFile.Create(owner);
  PuzzleStoneGfx[ptLine].LoadFromFile(path+'st-puzzle-ns.o3m');
  PuzzleStoneGfx[ptCross]:=TO3mFile.Create(owner);
  PuzzleStoneGfx[ptCross].LoadFromFile(path+'st-puzzle-nesw.o3m');
  PuzzleStoneGfx[ptHole]:=TO3mFile.Create(owner);
  PuzzleStoneGfx[ptHole].LoadFromFile(path+'st-puzzle-hole.o3m');
 end;

 procedure QuitLib;
 begin
  PuzzleStoneGfx[ptRound].Free;
  PuzzleStoneGfx[ptLine].Free;
  PuzzleStoneGfx[ptT].Free;
  PuzzleStoneGfx[ptEnd].Free;
  PuzzleStoneGfx[ptCross].Free;
  PuzzleStoneGfx[ptHole].Free;
 end;


  exports InitLib,
          QuitLib,
          MakeStone;
begin
end.
 
