unit o3editor; 

interface

uses
  Classes, SysUtils, o3menu, o3world, o3game, o3utils,o3objects,o3vectors,dglOpenGL,o3interfaces,o3actors,Math,o3basicobj,o3stones,o3item,StrUtils,o3levelinterface;

type TEditingType = (edStones, edItems, edActors, edTest, edSelect);
     TO3Editor = class;
     // ------------------ EDITOR MAIN MENU --------------------------
     TO3EditorMainMenu = class
      editor:TO3Editor;
      owner:TOzoneMainClass;
      menu:TO3Menu;
      constructor create(owner:TOzoneMainClass);
      destructor destroy; override;
      procedure run;
      procedure keyDown(sender:TO3Menu;key:TO3Key);
      procedure loadLevel(sender:TO3Control);
      procedure newLevel(sender:TO3Control);
      procedure exitMenu(sender:TO3Control);
     end;
     // ------------------- LEVEL LOAD/SAVE --------------------------
     TO3EditorLevelSelector = class
      edit:TOzoneMainClass;
      menu:TO3Menu;
      runlevel:integer;
      title:TO3CustomButton;
      nameI,packI:TO3TextEdit;
      constructor Create(owner:TOzoneMainClass);
      procedure OkClick(sender:TO3Control);
      procedure CancelClick(sender:TO3Control);
      function Execute(const title:string;var pack:string;var name:string; stdpack:string;stdname:string;omenu:TO3Menu=nil):boolean;
      destructor Destroy; override;
     end;
     // ------------------- SELECTION HELPER -------------------------
     TEditorSelection = class
      fList:TList;
      editor:TO3Editor;
      // ------- HELPER ROUTINES -------
      constructor Create(ed:TO3Editor);
      destructor Destroy; override;
      procedure highlight(o:TGameObject);
      procedure lowlight(o:TGameObject);

      // ------- SELECTION ROUTINES -------
      procedure select(o:TGameObject);
      procedure deselect(o:TGameObject);
      procedure deselectAll;

      // ------- HANDLING ROUTINES --------
      procedure deleteSelected;
      procedure moveSelected(dx,dy:single);
      procedure copyToBuffer;
     end;

     // ------------------- INPUT LISTENER -------------------------
     TEditorListener = class(TO3InputListener)
      editor:TO3Editor;
      modShift:boolean;
      modControl:boolean;
      mouseClicked:integer;
      ax,ay,mx,my:single;
      constructor Create(edit:TO3Editor);
      procedure mkeyDown(sender:TO3Menu;key:TO3Key);
      procedure mkeyUp(sender:TO3Menu;key:TO3Key);
      procedure mouseMove(xx,yy:integer); override;
      procedure mouseDown(button:integer); override;
      procedure mouseUp(button:integer); override;
     end;

     // ------------------- PARAMETER EDITOR --------------------------
     TO3ParamEditor=class
      owner:TOzoneMainClass;
      pList:TParameterList;
      m:TO3Menu;
      okLevel:integer;
      procedure OKClick(sender:TO3Control);
      procedure CancelClick(sender:TO3Control);
      procedure boolAct(sender:TO3Control);
      procedure stringAct(sender:TO3Control);
      procedure uintAct(sender:TO3Control);
      procedure intAct(sender:TO3Control);
      procedure floatAct(sender:TO3Control);
      procedure heightAct(sender:TO3Control);
      procedure posXAct(sender:TO3Control);
      procedure posZAct(sender:TO3Control);
      constructor create(own:TOzoneMainClass);
      destructor destroy; override;
      function Run(var o:TGameObject):boolean;
     end;
     
     // ------------------- LEVEL SIZE EDITOR -------------------------
     TLevelSizeSelector = class
      x,y,z:cardinal;
      rlev:integer;
      menu:TO3Menu;
      owner:TOzoneMainClass;
      procedure ActionX(sender:TO3Control);
      procedure ActionY(sender:TO3Control);
      procedure ActionZ(sender:TO3Control);
      procedure okClick(sender:TO3Control);
      procedure cancelClick(sender:TO3Control);
      function Execute:boolean;
      constructor Create(own:TOzoneMainClass);
      destructor Destroy; override;
     end;

     // ------------------- EDITOR MAIN CLASS -------------------------
     TO3Editor = class
      linking:boolean;
      linkSrc:TGameObject;
      okClickLevel:integer;
      curLev,curPack:string;
      listener:TEditorListener;
      selection:TEditorSelection;
      owner:TOzoneMainClass;
      menu:TO3Menu;
      doExit:boolean;
      levelChanged:boolean;
      cPos,cCenter:TVector3d; // camera position and view center
      ePlane:integer; // -1 overview; 0..theLevel.FY-1: editing plane;
      eType:TEditingType; // editing type
      buffer:TGameObject;
      levselect:TO3EditorLevelSelector;
      // buffer selector
      procedure processPaletteFiles;
      procedure processPaletteFile(name:string;sx,sy:single);
      procedure BufferSelectCallback(sender:TO3Control);
      procedure BufferSelectButtonClick(sender:TO3Control);
      // button toolbar
      procedure saveClick(sender:TO3Control);
      procedure stoneClick(sender:TO3Control);
      procedure itemClick(sender:TO3Control);
      procedure actorClick(sender:TO3Control);
      procedure testClick(sender:TO3Control);
      procedure planeupClick(sender:TO3Control);
      procedure planedownClick(sender:TO3Control);
      procedure copybufClick(sender:TO3Control);
      procedure bufparamClick(sender:TO3Control);
      procedure selectparamClick(sender:TO3Control);
      procedure linkClick(sender:TO3Control);
      // OK interface
      procedure initOKClick;
      procedure OKClick(sender:TO3Control);
      procedure CancelClick(sender:TO3Control);


      // ---------- HELPER ----------
      procedure ExitEditor(sender:TO3Control);
      procedure RunEditor;
      procedure updateEditorState;
      procedure tickEditorState;
      constructor Create(own:TOzoneMainClass);
      destructor Destroy; override;
      function mouseToLevWindow(var mx,my:single):boolean;
      function levWindowToLevPos(var mx,my:single):boolean;


      // ---------- HIDE/SHOW ------------
      procedure hideAllActors;
      procedure showAllActors;
      procedure hideAllItems;
      procedure showAllItems;
      procedure hideAllStones;
      procedure showAllStones;

      // ----------- HANDLING -----------
      procedure deleteObject(o:TGameObject; doFree:boolean=true);
      function moveObject(o:TGameObject;dx,dy,dz:single):boolean;
      procedure paramEdit(var o:TGameObject);
      procedure reCreate(var o:TGameObject);
     end;

implementation
uses o3m;

procedure TO3EditorLevelSelector.OkClick(sender:TO3Control);
begin
 runLevel:=1;
end;

procedure TO3EditorLevelSelector.CancelClick(sender:TO3Control);
begin
 runLevel:=-1;
end;

constructor TO3EditorLevelSelector.Create(owner:TOzoneMainClass);
var button:TO3CustomButton;
begin
 edit:=owner;
 menu:=TO3Menu.Create(edit);


 title:=TO3CustomButton.Create(edit.gfxSys,'titleD');
 title.BgModel:=nil;
 title.bgMtype:=plane;
 title.Text:='Title';
 title.x:=0;
 title.y:=0.75;
 title.H:=0.3;
 title.W:=0.5;
 title.TextSize:=1.5;
 menu.AddControl(title);

 button:=TO3CustomButton.Create(edit.gfxSys,'packD');
 button.BgModel:=nil;
 button.bgMtype:=plane;
 button.Text:='Level Pack:';
 button.x:=0;
 button.y:=0.50;
 button.H:=0.3;
 button.W:=0.5;
 button.TextSize:=1;
 menu.AddControl(button);

 packI:=TO3TextEdit.Create(edit.gfxSys,'packE');
 packI.X:=0;
 packI.Y:=0.4;
 menu.AddControl(packI);

 button:=TO3CustomButton.Create(edit.gfxSys,'nameD');
 button.BgModel:=nil;
 button.bgMtype:=plane;
 button.Text:='Level File Name:';
 button.x:=0;
 button.y:=0.10;
 button.H:=0.3;
 button.W:=0.5;
 button.TextSize:=1;
 menu.AddControl(button);

 nameI:=TO3TextEdit.Create(edit.gfxSys,'nameE');
 nameI.X:=0;
 nameI.Y:=0.0;
 menu.AddControl(nameI);

 button:=TO3CustomButton.Create(edit.gfxSys,'okB');
 button.bgModel:=edit.gfxSys.CreatePlaneXYModel(0.4,0.4,'button_ok.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=0.8;
 button.y:=-0.5;
 button.H:=0.4;
 button.W:=0.4;
 button.OnClick:=OkClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(edit.gfxSys,'cancelB');
 button.bgModel:=edit.gfxSys.CreatePlaneXYModel(0.4,0.4,'button_cancel.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=-0.8;
 button.y:=-0.5;
 button.H:=0.4;
 button.W:=0.4;
 button.OnClick:=CancelClick;
 menu.AddControl(button);

 edit.gfxSys.FreePlaneXYModel(menu.mouseModel); // Free the standard mouse model
 menu.mouseModel:=edit.gfxSys.CreatePlaneXYModel(0.3,0.3,'editor_mouse.png',BitmapSubDir); // load own mouse model

 menu.showToolTips:=false;
 menu.active:=false;
end;

function TO3EditorLevelSelector.Execute(const title:string;var pack:string;var name:string; stdpack:string;stdname:string;omenu:TO3Menu=nil):boolean;
begin
 self.title.Text:=title;
 packI.Text:=stdpack;
 nameI.Text:=stdname;
 if(omenu<>nil) then omenu.active:=false;
 runlevel:=0;
 menu.active:=true;
 while runLevel=0 do begin
  edit.InputSys.Tick;
  if edit.gfxSys.noclear then edit.gfxSys.clear;
  edit.gfxSys.setViewMatrix;
  edit.gfxSys.loadIdentityMatrix;
  edit.gfxSys.DoMenu(menu);
  if edit.gfxSys.noswap then edit.gfxSys.swap;
 end;
 if runlevel=1 then begin
  pack:=packI.Text;
  name:=nameI.Text;
  Result:=true;
 end
 else Result:=false;
 menu.active:=false;
 if(omenu<>nil) then omenu.active:=true;
end;

destructor TO3EditorLevelSelector.Destroy;
begin
 menu.Free;
end;


procedure TEditorSelection.highlight(o:TGameObject);
begin
 if o.gfxInfo=nil then o.gfxInfo:=newGfxInfo;
 o.gfxInfo.highlightColor[0]:=0;
 o.gfxInfo.highlightColor[1]:=1;
 o.gfxInfo.highlightColor[2]:=0;
 o.gfxInfo.highlightColor[3]:=1;
end;

procedure TEditorSelection.deleteSelected;
var x:integer;
begin
 if FList.Count<=0 then EXIT;
 for x:=0 to FList.Count-1 do
  editor.deleteObject(TGameObject(FList[x]));
 FList.Clear;
end;

procedure TEditorSelection.moveSelected;
var x:integer;
begin
 if FList.Count<=0 then EXIT;
 for x:=0 to FList.Count-1 do
  editor.moveObject(TGameObject(FList[x]),dx,0,dy);
end;

procedure TEditorSelection.copyToBuffer;
var t:TGameObject;
begin
 if(FList.Count<>1) then EXIT;
 if(editor.buffer<>nil) then editor.buffer.Free;
 t:=FList[0];
 if(t.objectType=otActor) then begin
  editor.buffer:=o3objects.CreateActor(t.Info['type'],-1,tActor(t).starty,-1,t.owner,t.Info['cmd']);
  tActor(editor.buffer).y:=-1;
  editor.buffer.Info['control']:=t.Info['control'];
 end;
 if(t.objectType=otItem) then begin
  editor.buffer:=o3objects.CreateItem(t.Info['cmd'],t.owner);
 end;
 if(t.objectType=otStone) then begin
  editor.buffer:=o3objects.CreateStone(t.Info['cmd'],t.owner);
 end;
end;

procedure TEditorSelection.lowlight(o:TGameObject);
begin
 if o.gfxInfo<>nil then begin
  o.gfxInfo.highlightColor[0]:=1;
  o.gfxInfo.highlightColor[1]:=1;
  o.gfxInfo.highlightColor[2]:=1;
 end;
end;

procedure TEditorSelection.select(o:TGameObject);
var sig:TSignal;
begin
 if editor.linking then begin
  if(o.objectType<>otItem)and(o.objectType<>otStone) then EXIT;
  if editor.linkSrc=nil then begin
   editor.linkSrc:=o;
   highLight(o);
  end
  else begin
   sig:=editor.owner.theLevel.signals.find(editor.linkSrc);
   if(sig<>nil) then sig.targets.Add(o)
   else begin
    sig:=TSignal.create;
    sig.source:=editor.linkSrc;
    sig.targets.Add(o);
    editor.owner.theLevel.signals.FList.Add(sig);
   end;
   lowLight(editor.linkSrc);
   editor.linkSrc:=nil;
   editor.linking:=false;
  end;
 end
 else begin
  fList.Remove(o);
  highLight(o);
  fList.Add(o);
 end;
end;

procedure TEditorSelection.deselect;
begin
 if editor.linking then EXIT;
 lowLight(o);
 fList.Remove(o);
end;

procedure TEditorSelection.deselectAll;
var x:integer;
begin
 if FList.Count<=0 then EXIT;
 for x:=0 to FList.Count-1 do
  lowlight(TGameObject(FList[x]));
 FList.Clear;
end;

constructor TEditorSelection.Create;
begin
 inherited Create;
 editor:=ed;
 FList:=TList.Create;
end;

destructor TEditorSelection.Destroy;
begin
 deselectAll;
 FList.Free;
 inherited Destroy;
end;


constructor TEditorListener.Create(edit:TO3Editor);
begin
 mouseClicked:=0;
 editor:=edit;
 modShift:=false;
end;

procedure TEditorListener.mkeyUp;
begin
 case key of
  O3K_LSHIFT:
  begin
   modShift:=false;
  end;
  O3K_LCTRL:
  begin
   modControl:=false;
  end;
 end;
end;

procedure TEditorListener.mouseUp(button:integer);
begin
 mouseClicked:=0;
end;

procedure TEditorListener.mouseMove;
var dx,dy:single;
    mxx,myy:single;
    cell:TCell;
    mxi,myi:integer;
    actor:TActor;
begin
 if mouseClicked=3 then begin
  if mx=-4E2 then EXIT;
  dx:=mx-ax;
  dy:=my-ay;
  editor.selection.moveSelected(-dx,-dy);
  mx:=editor.menu.mouseX;
  my:=editor.menu.mouseY;
  if (editor.mouseToLevWindow(mx,my) and editor.LevWindowToLevPos(mx,my)) then begin
   dx:=mx-ax;
   dy:=my-ay;
   editor.selection.moveSelected(dx,dy);
  end
  else mx:=-4E2;
 end;
 if (mouseClicked=1)and(editor.buffer<>nil)and(not editor.linking) then begin
 mxx:=editor.menu.mouseX;
 myy:=editor.menu.mouseY;
 if editor.mouseToLevWindow(mxx,myy) then begin
  editor.levelChanged:=true;
  if editor.LevWindowToLevPos(mxx,myy) then begin
   mxi:=floor(mxx);
   myi:=floor(myy);
   case editor.buffer.objectType of
    otActor: begin
     if editor.eType<>edActors then EXIT;
     mouseClicked:=0;
     actor:=o3objects.CreateActor(editor.buffer.Info['type'],mxx,TActor(editor.buffer).starty,myy,editor.buffer.owner,editor.buffer.Info['cmd']);
     actor.Info['control']:=editor.buffer.Info['control'];
     actor.name:='leved'+inttostr(editor.owner.theLevel.FActors.List.Count);
     actor.info['name']:=actor.name;
     editor.owner.theLevel.FActors.AddActor(actor,actor.name);
     if actor.Info['control']='firstplayer' then begin
      if editor.owner.theLevel.theActorToFollow=nil then
       editor.owner.theLevel.theActorToFollow:=actor
      else
       actor.Info['control']:='$firstplayer';
     end;
     if actor.Info['control']='secondplayer' then begin
      if editor.owner.theLevel.theActorToFollow=nil then
       editor.owner.theLevel.theActorToFollow:=actor
      else
       actor.Info['control']:='$secondplayer';
     end;
    end;
    otItem: begin
     if editor.eType<>edItems then EXIT;
     if editor.ePlane=-1 then EXIT;
     cell:=editor.owner.theLevel.getCell(mxi,editor.ePlane,myi);
     if cell<>nil then begin
      if(cell.FItem<>nil) then begin
       editor.selection.deselect(cell.FItem);
       Cell.FItem.Free;
      end;
      cell.FItem:=o3objects.CreateItem(editor.buffer.Info['cmd'],editor.buffer.owner);
      cell.FItem.x:=mxi;
      cell.FItem.y:=editor.ePlane;
      cell.FItem.z:=myi;
     end;
    end;
    otStone: begin
     if editor.eType<>edStones then EXIT;
     if editor.ePlane=-1 then EXIT;
     cell:=editor.owner.theLevel.getCell(mxi,editor.ePlane,myi);
     if cell<>nil then begin
      if(cell.FStone<>nil) then begin
       editor.selection.deselect(cell.FStone);
       Cell.FStone.Free;
      end;
      cell.FStone:=o3objects.CreateStone(editor.buffer.Info['cmd'],editor.buffer.owner);
      cell.FStone.x:=mxi;
      cell.FStone.y:=editor.ePlane;
      cell.FStone.z:=myi;
     end;
    end;

   end;
  end;
 end;
end;
end;

procedure TEditorListener.mouseDown(button:integer);
var mx,my:single;
    b:boolean;
    cell:TCell;
    x:integer;
    actor:TActor;
begin
mouseClicked:=button;
ax:=editor.menu.mouseX;
ay:=editor.menu.mouseY;
if (editor.mouseToLevWindow(ax,ay) and editor.LevWindowToLevPos(ax,ay)) then begin
 self.mx:=ax;
 self.my:=ay;
end
else self.mx:=-4E2;
if (button=3)or(editor.linking) then begin
 mx:=editor.menu.mouseX;
 my:=editor.menu.mouseY;
 if editor.mouseToLevWindow(mx,my) then begin
  editor.levelChanged:=true;
  b:=editor.LevWindowToLevPos(mx,my);
 // editor.menu.aToolTip:='mx: '+copy(FloatToStrF(mx,ffFixed,4,5),1,5)+' my: '+copy(FloatToStrF(my,ffFixed,4,5),1,5)+' l: '+booltostr(b);

  if b and (editor.ePlane<>-1)and(editor.eType=edStones) then begin
   cell:=editor.owner.theLevel.getCell(floor(mx),editor.ePlane,floor(my));
   if (cell<>nil)and(cell.FStone<>nil) then begin
    if not modShift then editor.selection.deselectAll;
    editor.selection.select(cell.FStone);
   end;
  end;
  if b and (editor.ePlane<>-1)and(editor.eType=edItems) then begin
   cell:=editor.owner.theLevel.getCell(floor(mx),editor.ePlane,floor(my));
   if (cell<>nil)and(cell.FItem<>nil) then begin
    if not modShift then editor.selection.deselectAll;
    editor.selection.select(cell.FItem);
   end;
  end;
  if b and (editor.eType=edActors) then begin
   if editor.owner.theLevel.FActors.List.Count>0 then begin
    for x:=0 to editor.owner.theLevel.FActors.List.Count-1 do begin
     actor:=editor.owner.theLevel.FActors.List[x];
     if actor<>nil then begin
      if(abs(actor.x-mx)<0.35)and(abs(actor.z-my)<0.35) then begin
       if not modShift then editor.selection.deselectAll;
       editor.selection.select(actor);
      end;
     end;
    end;
   end;
  end;
 end;
end
else mouseMove(0,0);
{else if (button=1)and(editor.buffer<>nil) then begin
 mx:=editor.menu.mouseX;
 my:=editor.menu.mouseY;
 if editor.mouseToLevWindow(mx,my) then begin
  editor.levelChanged:=true;
  if editor.LevWindowToLevPos(mx,my) then begin
   mxi:=floor(mx);
   myi:=floor(my);
   case editor.buffer.objectType of
    otActor: begin
     if editor.eType<>edActors then EXIT;
     actor:=o3objects.CreateActor(editor.buffer.Info['type'],mx,TActor(editor.buffer).starty,my,editor.buffer.owner,editor.buffer.Info['cmd']);
     actor.Info['control']:=editor.buffer.Info['control'];
     actor.name:='leved'+inttostr(editor.owner.theLevel.FActors.List.Count);
     actor.info['name']:=actor.name;
     editor.owner.theLevel.FActors.AddActor(actor,actor.name);
     if actor.Info['control']='firstplayer' then begin
      if editor.owner.theLevel.theActorToFollow=nil then
       editor.owner.theLevel.theActorToFollow:=actor
      else
       actor.Info['control']:='$firstplayer';
     end;
     if actor.Info['control']='secondplayer' then begin
      if editor.owner.theLevel.theActorToFollow=nil then
       editor.owner.theLevel.theActorToFollow:=actor
      else
       actor.Info['control']:='$secondplayer';
     end;
    end;
    otItem: begin
     if editor.eType<>edItems then EXIT;
     if editor.ePlane=-1 then EXIT;
     cell:=editor.owner.theLevel.getCell(mxi,editor.ePlane,myi);
     if cell<>nil then begin
      if(cell.FItem<>nil) then begin
       editor.selection.deselect(cell.FItem);
       Cell.FItem.Free;
      end;
      cell.FItem:=o3objects.CreateItem(editor.buffer.Info['cmd'],editor.buffer.owner);
      cell.FItem.x:=mxi;
      cell.FItem.y:=editor.ePlane;
      cell.FItem.z:=myi;
     end;
    end;
    otStone: begin
     if editor.eType<>edStones then EXIT;
     if editor.ePlane=-1 then EXIT;
     cell:=editor.owner.theLevel.getCell(mxi,editor.ePlane,myi);
     if cell<>nil then begin
      if(cell.FStone<>nil) then begin
       editor.selection.deselect(cell.FStone);
       Cell.FStone.Free;
      end;
      cell.FStone:=o3objects.CreateStone(editor.buffer.Info['cmd'],editor.buffer.owner);
      cell.FStone.x:=mxi;
      cell.FStone.y:=editor.ePlane;
      cell.FStone.z:=myi;
     end;
    end;

   end;
  end;
 end;
end;}
end;

procedure TEditorListener.mkeyDown;
begin
 if editor.eType<>edSelect then begin
  case key of
   O3K_LSHIFT:
   begin
    modShift:=true;
   end;
   O3K_LCTRL:
   begin
    modControl:=true;
   end;


   O3K_DELETE:
   begin
    editor.selection.deleteSelected;
    editor.levelChanged:=true;
   end;


   O3K_ESCAPE:
   begin
    if(editor.linking) then begin
     editor.linking:=false;
     if(editor.linkSrc<>nil) then
      editor.selection.lowlight(editor.linkSrc);
     editor.linkSrc:=nil;
    end;
   end;

   O3K_A:
   begin
    editor.actorClick(nil);
   end;
   O3K_I:
   begin
    editor.itemClick(nil);
   end;
   O3K_S:
   begin
    editor.stoneClick(nil);
   end;
   O3K_C:
   begin
    editor.copybufClick(nil);
   end;
   O3K_T:
   begin
    editor.testClick(nil);
   end;
   O3K_W:
   begin
    editor.saveClick(nil);
   end;
   O3K_P:
   begin
    editor.selectparamClick(nil);
   end;
   O3K_V:
   begin
//    editorsaveClick(nil); //LEVEL attribs
   end;
   O3K_B:
   begin
    editor.bufparamClick(nil);
   end;
   O3K_L:
   begin
    editor.linkClick(nil);
   end;
{   O3K_R:
   begin
    if editor.levselect.Execute(p,n,editor.curPack,editor.curLev) then
     if FileExists(editor.owner.path+LevelSubDir+p+PathDelim+n+'.o3') then begin
//      editor.owner.theLevel.saveToFile(editor.owner.path+LevelSubDir+p+PathDelim+n+'.o3','');
      editor.curLev:=n;
      editor.curPack:=p;
      editor.selection.deselectAll;
      editor.buffer.Free;
      if editor.owner.theLevel<>nil then editor.owner.theLevel.Free;
      editor.owner.theLevel:=TLevel.Create(editor.owner,editor.owner.path+LevelSubDir+'import_ozone.so',editor.owner.path+LevelSubDir+editor.curPack+PathDelim+editor.curLev+'.o3','nil',CreateStone,CreateItem,CreateActor);
     end;
   end;
 }




   O3K_PageUp:
   begin
    editor.planeupClick(nil);
   end;
   O3K_PageDown:
   begin
    editor.planedownClick(nil);
   end;
   O3K_LEFT:
   begin
    editor.cCenter[0]:=editor.cCenter[0]+1;
    editor.cPos[0]:=editor.cPos[0]+1;
   end;
   O3K_RIGHT:
   begin
    editor.cCenter[0]:=editor.cCenter[0]-1;
    editor.cPos[0]:=editor.cPos[0]-1;
   end;
   O3K_UP:
   begin
    editor.cCenter[2]:=editor.cCenter[2]+1;
    editor.cPos[2]:=editor.cPos[2]+1;
   end;
   O3K_DOWN:
   begin
    editor.cCenter[2]:=editor.cCenter[2]-1;
    editor.cPos[2]:=editor.cPos[2]-1;
   end;

  end;
 end
 else begin
  if key=O3K_ESCAPE then
   editor.eType:=edStones;
 end;
end;




procedure TO3Editor.processPaletteFiles;
begin
 processPaletteFile('index.eac',0.84,0.5);
 processPaletteFile('index.est',1.04,0.5);
 processPaletteFile('index.eit',1.24,0.5);
end;

procedure TO3Editor.processPaletteFile(name:string;sx,sy:single);
var f:TextFile;
    s:string;
    fn,img,desc:string;
    b:TO3CustomButton;
begin
 AssignFile(f,owner.path+EditorSubDir+name);
 Reset(f);
 while not EOF(f) do begin
  Readln(f,s);
  if(s='')or(s[1]=';') then CONTINUE;
  fn:=getToken(s,1,':');
  img:=getToken(s,2,':');
  desc:=getToken(s,3,':');
  b:=TO3CustomButton.Create(owner.gfxSys,fn);
  b.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.15,0.15,img,EditorSubDir);
  b.bgMtype:=plane;
  b.x:=sx;
  b.y:=sy;
  b.H:=0.15;
  b.W:=0.15;
  b.ToolTip:='Select: '+desc;
  b.OnClick:=BufferSelectCallback;
  menu.AddControl(b);
  sy:=sy-0.2;
 end;
 CloseFile(f);
end;

procedure TO3Editor.BufferSelectButtonClick(sender:TO3Control);
begin
 if (GetToken(sender.name,2,'|')='nil') then exit;
 if buffer<>nil then buffer.Free;
 if LeftStr(GetToken(sender.name,2,'|'),2)='st' then begin
  buffer:=CreateStone(GetToken(sender.name,2,'|'),owner.theLevel);
 end
 else if LeftStr(GetToken(sender.name,2,'|'),2)='it' then begin
  buffer:=CreateItem(GetToken(sender.name,2,'|'),owner.theLevel);
 end
 else if LeftStr(GetToken(sender.name,2,'|'),2)='ac' then begin
  buffer:=o3objects.CreateActor(getToken(GetToken(sender.name,2,'|'),1,';'),-1,3,-1,owner.theLevel,getToken(GetToken(sender.name,2,'|'),3,';'));
  TActor(buffer).y:=-1;
  buffer.info['control']:=getToken(GetToken(sender.name,2,'|'),2,';');
 end;
 eType:=edStones;
end;


procedure TO3Editor.BufferSelectCallback(sender:TO3Control);
var a:array of array of string;
    b:array of array of TGameObject;
    t:TextFile;
    x,y,i:integer;
    s,r:string;
    oldType:TEditingType;
    theMenu:TO3Menu;
    theButton:TO3CustomButton;
    delta,ts:cardinal;
begin
 if not FileExists(owner.path+EditorSubDir+sender.name) then EXIT;
 AssignFile(t,owner.path+EditorSubDir+sender.name);
 Reset(t);
 y:=0;
 while not Eof(t) do begin
  Readln(t,s);
  if s='' then CONTINUE;
  setLength(a,y+1);
  x:=0;
  i:=1;
  r:='';
  while i<=Length(s) do begin
   if(s[i]=',') then begin
    setLength(a[y],x+1);
    a[y,x]:=r;
    inc(x);
    r:='';
   end
   else if s[i]='''' then r:=r+','
   else if s[i]<>' ' then r:=r+s[i];
   inc(i);
  end;
  if r<>'' then begin
   setLength(a[y],x+1);
   a[y,x]:=r;
   r:='';
  end;
  inc(y);
 end;
 CloseFile(t);
 if Length(a)<=0 then EXIT;

 oldType:=eType;
 eType:=edSelect;
 menu.active:=false;
 theMenu:=TO3Menu.Create(owner);
 owner.gfxSys.FreePlaneXYModel(theMenu.mouseModel); // Free the standard mouse model
 theMenu.mouseModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'editor_mouse.png',BitmapSubDir); // load own mouse model
 theMenu.showToolTips:=false;
 theMenu.active:=true;

 setLength(b,Length(a));

 for y:=0 to Length(a)-1 do begin
  setLength(b[y],Length(a[y]));
  if Length(a[y])<=0 then CONTINUE;
  for x:=0 to Length(a[y])-1 do begin
   if LeftStr(a[y,x],2)='st' then begin
    b[y,x]:=CreateStone(a[y,x],owner.theLevel);
   end
   else if LeftStr(a[y,x],2)='it' then begin
    b[y,x]:=CreateItem(a[y,x],owner.theLevel);
   end
   else if LeftStr(a[y,x],2)='ac' then begin
    b[y,x]:=o3objects.CreateActor(getToken(a[y,x],1,';'),-1,-1,-1,owner.theLevel,getToken(a[y,x],3,';'));
    b[y,x].info['control']:=getToken(a[y,x],2,';');
   end;
   theButton:=TO3CustomButton.Create(owner.gfxSys,inttostr(x)+'x'+inttostr(y)+'|'+a[y,x]);
   theButton.X:=-1.075+0.3*x;
   theButton.Y:=0.75-0.3*y;
   theButton.W:=0.3;
   theButton.H:=0.3;
   theButton.OnClick:=BufferSelectButtonClick;
   theMenu.AddControl(theButton);
  end;
 end;

 ts:=owner.IOSys.getTicks;
 while eType=edSelect do begin
  delta:=owner.IOSys.getTicks-ts;
  ts:=owner.IOSys.getTicks;
  owner.gfxSys.clear;
  owner.InputSys.Tick;
  owner.gfxSys.setProjectionMatrix;
  owner.gfxSys.pushmatrix;
  owner.gfxSys.loadIdentityMatrix;
  owner.gfxSys.setPerspective(70,1.33,0.1,100);
  owner.gfxSys.setViewMatrix;
  owner.gfxSys.loadIdentityMatrix;
  for y:=0 to Length(b)-1 do begin
   if Length(b[y])<=0 then CONTINUE;
   for x:=0 to Length(b[y])-1 do begin
    if b[y,x]=nil then CONTINUE;
    owner.gfxSys.viewport(-1.5+x*0.3,-0.4+y*0.3,1.65,1.2);
    owner.gfxSys.pushmatrix;
    owner.gfxSys.lookAt(0,6,0,0,0,0,0,1,1);
    owner.gfxSys.setLightPos(0,6,0,1);
    owner.gfxSys.renderPass:=0;
    owner.gfxSys.translate(2,0,-2);
    if b[y,x].objectType=otActor then
     owner.gfxSys.translate(1.5,2,1.5);
    b[y,x].Render(delta);
    owner.gfxSys.renderPass:=1;
    b[y,x].Render(delta);
    owner.gfxSys.popmatrix;
   end;
  end;
  owner.gfxSys.setProjectionMatrix;
  owner.gfxSys.popmatrix;
  owner.gfxSys.viewport(-1.33,1,2.66,2);
  owner.gfxSys.setViewMatrix;
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  owner.gfxSys.DoMenu(theMenu);
  glPopAttrib;
  owner.gfxSys.swap;
 end;
 for y:=0 to Length(b)-1 do begin
  if Length(b[y])<=0 then CONTINUE;
  for x:=0 to Length(b[y])-1 do begin
   if b[y,x]=nil then CONTINUE;
   b[y,x].Free;
  end;
 end;
 finalize(a);
 finalize(b);
 eType:=oldType;
 menu.active:=true;
 theMenu.Free;
end;

procedure TO3Editor.reCreate(var o:TGameObject);
var t:TGameObject;
    cell:TCell;
begin
 case o.objectType of
  otActor: begin
   t:=o3objects.CreateActor(o.Info['type'],TActor(o).startx,TActor(o).starty,TActor(o).startz,o.owner,o.Info['cmd']);
   TActor(t).x:=TActor(o).x;
   TActor(t).y:=TActor(o).y;
   TActor(t).z:=TActor(o).z;
   TActor(t).Info['control']:=TActor(o).Info['control'];
   if selection.fList.IndexOf(o)>-1 then begin
    selection.deselect(o);
    selection.select(t);
   end;
    TActor(t).name:=TActor(o).name;//'actor'+inttostr(owner.theLevel.FActors.List.Count);
    t.Info['name']:=TActor(t).name;
   
   if owner.theLevel.FActors.List.IndexOf(o)>-1 then begin
    TActor(t).y:=TActor(t).starty;
    owner.theLevel.FActors.RemoveActor(TActor(o));
//    t.Info['name']:='actor'+inttostr(owner.theLevel.FActors.List.Count);
    owner.theLevel.FActors.AddActor(TActor(t),TActor(t).name);
   end;
   if owner.theLevel.theActorToFollow=o then
    owner.theLevel.theActorToFollow:=TActor(t);
   if owner.theLevel.theActorToFollow2=o then
    owner.theLevel.theActorToFollow2:=TActor(t);
   o.Free;
   o:=t;
  end;
  otItem: begin
   t:=o3objects.CreateItem(o.Info['cmd'],o.owner);
   TItem(t).x:=TItem(o).x;
   TItem(t).y:=TItem(o).y;
   TItem(t).z:=TItem(o).z;
   cell:=owner.theLevel.getCell(TItem(o).x,TItem(o).y,TItem(o).z);
   if cell<>nil then begin
    cell.FItem:=TItem(t);
    cell.FItem:=TItem(t);
    cell.FItem:=TItem(t);
   end;
   if selection.fList.IndexOf(o)>-1 then begin
    selection.deselect(o);
    selection.select(t);
   end;
   o.Free;
   o:=t;
  end;
  otStone: begin
   t:=o3objects.CreateStone(o.Info['cmd'],o.owner);
   TStone(t).x:=TStone(o).x;
   TStone(t).y:=TStone(o).y;
   TStone(t).z:=TStone(o).z;
   cell:=owner.theLevel.getCell(TStone(o).x,TStone(o).y,TStone(o).z);
   if cell<>nil then begin
    cell.FStone:=TStone(t);
    cell.FStone:=TStone(t);
    cell.FStone:=TStone(t);
   end;
   if selection.fList.IndexOf(o)>-1 then begin
    selection.deselect(o);
    selection.select(t);
   end;
   o.Free;
   o:=t;
  end;
 end;
end;

procedure TO3Editor.initOKClick;
begin
 OkClickLevel:=0;
end;

procedure TO3Editor.OKClick;
begin
 OkClickLevel:=1;
end;

procedure TO3Editor.CancelClick;
begin
 OkClickLevel:=-1;
end;

procedure TO3ParamEditor.heightAct(sender:TO3Control);
begin
 DecimalSeparator:='.';
 pList[strToInt(sender.name)].heightValue:=StrToFloat(TO3TextControl(sender).Text);
end;

procedure TO3ParamEditor.posXAct(sender:TO3Control);
begin
 DecimalSeparator:='.';
 PPosParam(pList[strToInt(sender.name)].pValue)^[0]:=StrToFloat(TO3TextControl(sender).Text);
end;

procedure TO3ParamEditor.posZAct(sender:TO3Control);
begin
 DecimalSeparator:='.';
 PPosParam(pList[strToInt(sender.name)].pValue)^[1]:=StrToFloat(TO3TextControl(sender).Text);
end;

procedure TO3ParamEditor.uintAct(sender:TO3Control);
begin
 pList[strToInt(sender.name)].uintValue:=StrToInt(TO3TextControl(sender).Text);
end;

procedure TO3ParamEditor.intAct(sender:TO3Control);
begin
 pList[strToInt(sender.name)].intValue:=StrToInt(TO3TextControl(sender).Text);
end;

procedure TO3ParamEditor.floatAct(sender:TO3Control);
begin
 DecimalSeparator:='.';
 pList[strToInt(sender.name)].floatValue:=StrToFloat(TO3TextControl(sender).Text);
end;

procedure TO3ParamEditor.stringAct(sender:TO3Control);
begin
 pList[strToInt(sender.name)].stringValue:=TO3TextControl(sender).Text;
end;

procedure TO3ParamEditor.boolAct(sender:TO3Control);
begin
 pList[strToInt(sender.name)].boolValue:=TO3CheckBox(sender).value;
end;

procedure TO3ParamEditor.OkClick(sender:TO3Control);
begin
 okLevel:=1;
end;

procedure TO3ParamEditor.CancelClick(sender:TO3Control);
begin
 okLevel:=-1;
end;

constructor TO3ParamEditor.create(own:TOzoneMainClass);
begin
 owner:=own;
 okLevel:=0;
end;

destructor TO3ParamEditor.destroy;
begin
 inherited Destroy;
end;

function TO3ParamEditor.Run(var o:TGameObject):boolean;
var x:integer;
    b:TO3CustomButton;
    e:TO3TextEdit;
    c:TO3CheckBox;
    y:single;

procedure addText(xx,yy:single;text:string);
begin
 b:=TO3CustomButton.Create(owner.gfxSys,floatToStr(xx)+'x'+floatToStr(yy)+'L');
 b.Text:=text;
 b.W:=2;
 b.H:=0.2;
 b.X:=xx;
 b.Y:=yy;
 b.TextSize:=0.8;
 b.WordWarp:=true;
 b.OnClick:=nil;
 m.AddControl(b);
 if yy=y then y:=y-0.1;
end;

begin
 Result:=false;
 m:=TO3Menu.Create(owner);

 b:=TO3CustomButton.Create(owner.gfxSys,'okB');
 b.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'button_ok.png',BitmapSubDir);
 b.bgMtype:=plane;
 b.x:=0.8;
 b.y:=-0.8;
 b.H:=0.3;
 b.W:=0.3;
 b.OnClick:=OkClick;
 m.AddControl(b);

 b:=TO3CustomButton.Create(owner.gfxSys,'cancelB');
 b.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'button_cancel.png',BitmapSubDir);
 b.bgMtype:=plane;
 b.x:=-0.8;
 b.y:=-0.8;
 b.H:=0.3;
 b.W:=0.3;
 b.OnClick:=CancelClick;
 m.AddControl(b);

 m.showToolTips:=false;

 owner.gfxSys.FreePlaneXYModel(m.mouseModel); // Free the standard mouse model
 m.mouseModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'editor_mouse.png',BitmapSubDir); // load own mouse model



 y:=0.9;
 pList:=o.getParameterList;
 if PList=nil then exit;
 if pList.Count>0 then begin
  for x:=0 to pList.Count-1 do begin
   case pList[x].pType of
    ptFloat://           addText(0,y,floatToStr(pList[x].floatValue));//writeln(pList[x].desc,' = ',pList[x].floatValue);
    begin
     addText(0,y,pList[x].desc);
     e:=TO3FloatEdit.Create(owner.gfxSys,inttostr(x));
     e.Text:=FloatToStr(pList[x].floatValue);
     e.X:=0;
     e.Y:=y;
     y:=y-0.15;
     e.OnClick:=nil;
     e.OnAction:=floatAct;
     m.AddControl(e);
    end;
    ptHeight://          addText(0,y,floatToStr(pList[x].heightValue));//writeln(pList[x].desc,' = ',pList[x].heightValue);
    begin
     addText(0,y,pList[x].desc);
     e:=TO3FloatEdit.Create(owner.gfxSys,inttostr(x));
     e.Text:=FloatToStr(pList[x].heightValue);
     e.X:=0;
     e.Y:=y;
     y:=y-0.15;
     e.OnClick:=nil;
     e.OnAction:=heightAct;
     m.AddControl(e);
    end;
    ptPos://             addText(0,y,'N/A');//writeln(pList[x].desc,' = ',pList[x].posValue);
    begin
     addText(0,y,pList[x].desc);
     b:=TO3CustomButton.Create(owner.gfxSys,inttostr(x)+'Xd');
     b.Text:='X:';
     b.W:=0.2;
     b.H:=0.2;
     b.X:=-1.17;
     b.Y:=y-0.03;
     b.TextSize:=1;
     b.OnClick:=nil;
     m.AddControl(b);

     b:=TO3CustomButton.Create(owner.gfxSys,inttostr(x)+'Yd');
     b.Text:='Z:';
     b.W:=2;
     b.H:=0.2;
     b.X:=0.03;
     b.Y:=y-0.03;
     b.TextSize:=1;
     b.OnClick:=nil;
     m.AddControl(b);


     e:=TO3FloatEdit.Create(owner.gfxSys,inttostr(x));
     e.Text:=FloatToStr(pList[x].posValue[0]);
     e.X:=-0.6;
     e.Y:=y;
     e.OnClick:=nil;
     e.OnAction:=posXact;
     m.AddControl(e);


     e:=TO3FloatEdit.Create(owner.gfxSys,inttostr(x));
     e.Text:=FloatToStr(pList[x].posValue[1]);
     e.X:=0.6;
     e.Y:=y;
     e.OnClick:=nil;
     e.OnAction:=posZact;
     m.AddControl(e);

     y:=y-0.15;
    end;
    ptInt:
    begin
     addText(0,y,pList[x].desc);
     addText(0,y,inttostr(pList[x].intValue));
    end; //writeln(pList[x].desc,' = ',pList[x].intValue);
    ptUInt:
    begin
     addText(0,y,pList[x].desc);
     e:=TO3UIntEdit.Create(owner.gfxSys,inttostr(x));
     e.Text:=IntToStr(pList[x].uintValue);
     e.X:=0;
     e.Y:=y;
     y:=y-0.15;
     e.OnClick:=nil;
     e.OnAction:=uintAct;
     m.AddControl(e);
           // addText(0,y,inttostr(pList[x].uintValue));//writeln(pList[x].desc,' = ',pList[x].uintValue);
    end;
    ptBool:
    begin
     c:=TO3CheckBox.Create(owner.gfxSys,inttostr(x));
     c.Text:=pList[x].desc;
     c.value:=pList[x].boolValue;
     c.X:=0.15;
     c.Y:=y;
     c.H:=0.1;
     c.W:=1.5;
     y:=y-0.15;
     c.OnAction:=boolAct;
     m.AddControl(c);
//     addText(0,y,pList[x].desc);
//     addText(0,y,boolToStr(pList[x].boolValue,true));//writeln(pList[x].desc,' = ',pList[x].boolValue);
    end;
    ptMulti:
    begin
     addText(0,y,pList[x].desc);
     addText(0,y,'N/A');//writeln(pList[x].desc,' = ',pList[x].multichoiceValue);
    end;
    ptString:
    begin
     addText(0,y,pList[x].desc);
     e:=TO3RichEdit.Create(owner.gfxSys,inttostr(x));
     e.Text:=pList[x].stringValue;
     e.X:=0;
     e.Y:=y-0.15;
     y:=y-0.35;
     e.OnClick:=nil;
     e.OnAction:=stringAct;
     m.AddControl(e);
    end;
   end;
   y:=y-0.1;
  end;
 end;
 m.active:=true;
 while (okLevel=0)and(not m.exitMenu) do begin
  owner.InputSys.Tick;
  owner.gfxSys.clear;
  owner.gfxSys.DoMenu(m);
  owner.gfxSys.swap;
 end;
 m.active:=false;
 if okLevel=1 then begin
  o.processParameterList(pList);
  Result:=true;
 end;
 pList.Free;
 m.Free;
end;


procedure TO3Editor.paramEdit(var o:TGameObject);
var pEd:TO3ParamEditor;
begin
 pEd:=TO3ParamEditor.create(owner);
 menu.active:=false;
 if pEd.Run(o) then reCreate(o);
 menu.active:=true;
 pEd.Free;
end;

procedure TO3Editor.deleteObject;
var cell:TCell;
    i:integer;
begin
 if owner.theLevel.signals.FList.Count>0 then begin
  i:=0;
  while i<owner.theLevel.signals.FList.Count do begin
   if TSignal(owner.theLevel.signals.FList[i]).source=o then begin
    TSignal(owner.theLevel.signals.FList[i]).Free;
    owner.theLevel.signals.FList[i]:=nil;
    owner.theLevel.signals.FList.Delete(i);
    CONTINUE;
   end;
   TSignal(owner.theLevel.signals.FList[i]).targets.Remove(o);
   inc(i);
  end;
 end;
 if o.objectType=otActor then begin
  owner.theLevel.FActors.List.Remove(o);
  if owner.theLevel.theActorToFollow=o then
   owner.theLevel.theActorToFollow:=nil;
  if owner.theLevel.theActorToFollow2=o then
   owner.theLevel.theActorToFollow2:=nil;
 end;
 if o.objectType=otStone then begin
  cell:=owner.theLevel.getCell(TStone(o).x,TStone(o).y,TStone(o).z);
  if(cell<>nil) then
   cell.FStone:=nil;
 end;
 if o.objectType=otItem then begin
  cell:=owner.theLevel.getCell(TItem(o).x,TItem(o).y,TItem(o).z);
  if(cell<>nil) then
   cell.FItem:=nil;
 end;
 if doFree then o.Free;
end;

function TO3Editor.moveObject;
var cell:TCell;
    dxi,dzi:integer;
begin
 Result:=false;
 if o.objectType=otActor then begin
  TActor(o).x:=TActor(o).x+dx;
  TActor(o).y:=TActor(o).y+dy;
  TActor(o).z:=TActor(o).z+dz;
  Result:=true;
 end;
 if o.objectType=otStone then begin
  dxi:=round(dx);
  dzi:=round(dz);
  cell:=owner.theLevel.getCell(TStone(o).x+dxi,TStone(o).y,TStone(o).z+dzi);
  if(cell<>nil)and(cell.FStone=nil) then
   cell.FStone:=TStone(o)
  else
    EXIT;
  cell:=owner.theLevel.getCell(TStone(o).x,TStone(o).y,TStone(o).z);
  if(cell<>nil) then
   cell.FStone:=nil;
  TStone(o).x:=TStone(o).x+dxi;
  TStone(o).z:=TStone(o).z+dzi;
  Result:=true;
 end;
 if o.objectType=otItem then begin
  dxi:=round(dx);
  dzi:=round(dz);
  cell:=owner.theLevel.getCell(TItem(o).x+dxi,TItem(o).y,TItem(o).z+dzi);
  if(cell<>nil)and(cell.FItem=nil) then
   cell.FItem:=TItem(o)
  else
    EXIT;
  cell:=owner.theLevel.getCell(TItem(o).x,TItem(o).y,TItem(o).z);
  if(cell<>nil) then
   cell.FItem:=nil;
  TItem(o).x:=TItem(o).x+dxi;
  TItem(o).z:=TItem(o).z+dzi;
  Result:=true;
 end;
end;


procedure TO3Editor.hideAllActors;
var x:integer;
begin
 if owner.theLevel.FActors.List.Count>0 then
  for x:=0 to owner.theLevel.FActors.List.Count-1 do begin
    if TActor(owner.theLevel.FActors.List[x]).gfxInfo=nil then
     TActor(owner.theLevel.FActors.List[x]).gfxInfo:=NewGfxInfo;
    if TActor(owner.theLevel.FActors.List[x]).gfxInfo.highlightColor[0]=-1 then begin
     TActor(owner.theLevel.FActors.List[x]).gfxInfo.highlightColor[0]:=1;
     TActor(owner.theLevel.FActors.List[x]).gfxInfo.highlightColor[1]:=1;
     TActor(owner.theLevel.FActors.List[x]).gfxInfo.highlightColor[2]:=1;
    end;
    TActor(owner.theLevel.FActors.List[x]).gfxInfo.highlightColor[3]:=0.4;
  end;
end;

procedure TO3Editor.showAllActors;
var x:integer;
begin
 if owner.theLevel.FActors.List.Count>0 then
  for x:=0 to owner.theLevel.FActors.List.Count-1 do begin
   if TActor(owner.theLevel.FActors.List[x]).gfxInfo<>nil then
    TActor(owner.theLevel.FActors.List[x]).gfxInfo.highlightColor[3]:=1;
  end;
end;

procedure TO3Editor.hideAllItems;
var x,y,z:integer;
    cell:TCell;
begin
 if owner.theLevel.FX<=0 then exit;
 if owner.theLevel.FY<=0 then exit;
 if owner.theLevel.FZ<=0 then exit;
 for x:=0 to owner.theLevel.FX-1 do
  for y:=0 to owner.theLevel.FY-1 do
   for z:=0 to owner.theLevel.FZ-1 do begin
    cell:=owner.theLevel.getCell(x,y,z);
    if (cell<>nil)and(cell.FItem<>nil) then begin
     if cell.FItem.gfxInfo=nil then cell.FItem.gfxInfo:=newGfxInfo;
     if cell.FItem.gfxInfo.highlightColor[0]=-1 then begin
      cell.FItem.gfxInfo.highlightColor[0]:=1;
      cell.FItem.gfxInfo.highlightColor[1]:=1;
      cell.FItem.gfxInfo.highlightColor[2]:=1;
     end;
     cell.FItem.gfxInfo.highlightColor[3]:=0.5;
    end;
   end;
end;

procedure TO3Editor.showAllItems;
var x,y,z:integer;
    cell:TCell;
begin
 if owner.theLevel.FX<=0 then exit;
 if owner.theLevel.FY<=0 then exit;
 if owner.theLevel.FZ<=0 then exit;
 for x:=0 to owner.theLevel.FX-1 do
  for y:=0 to owner.theLevel.FY-1 do
   for z:=0 to owner.theLevel.FZ-1 do begin
    cell:=owner.theLevel.getCell(x,y,z);
    if (cell<>nil)and(cell.FItem<>nil) then
     if cell.FItem.gfxInfo<>nil then
      cell.FItem.gfxInfo.highlightColor[3]:=1;
   end;
end;

procedure TO3Editor.showAllStones;
var x,y,z:integer;
    cell:TCell;
begin
 if owner.theLevel.FX<=0 then exit;
 if owner.theLevel.FY<=0 then exit;
 if owner.theLevel.FZ<=0 then exit;
 for x:=0 to owner.theLevel.FX-1 do
  for y:=0 to owner.theLevel.FY-1 do
   for z:=0 to owner.theLevel.FZ-1 do begin
    cell:=owner.theLevel.getCell(x,y,z);
    if (cell<>nil)and(cell.FStone<>nil) then
     if cell.FStone.gfxInfo<>nil then
      cell.FStone.gfxInfo.highlightColor[3]:=1;
   end;
end;

procedure TO3Editor.hideAllStones;
var x,y,z:integer;
    cell:TCell;
begin
 if owner.theLevel.FX<=0 then exit;
 if owner.theLevel.FY<=0 then exit;
 if owner.theLevel.FZ<=0 then exit;
 for x:=0 to owner.theLevel.FX-1 do
  for y:=0 to owner.theLevel.FY-1 do
   for z:=0 to owner.theLevel.FZ-1 do begin
    cell:=owner.theLevel.getCell(x,y,z);
    if (cell<>nil)and(cell.FStone<>nil) then begin
     if cell.FStone.gfxInfo=nil then cell.FStone.gfxInfo:=newGfxInfo;
     if cell.FStone.gfxInfo.highlightColor[0]=-1 then begin
      cell.FStone.gfxInfo.highlightColor[0]:=1;
      cell.FStone.gfxInfo.highlightColor[1]:=1;
      cell.FStone.gfxInfo.highlightColor[2]:=1;
     end;
     cell.FStone.gfxInfo.highlightColor[3]:=0.5;
    end;
   end;
end;











procedure TO3Editor.updateEditorState;
begin
 selection.deselectAll;
 if eType=edStones then begin
  hideAllActors;
  hideAllItems;
  showAllStones;
  ePlane:=-1;
 end;
 if eType=edTest then begin
  showAllStones;
  showAllItems;
  showAllActors;
  ePlane:=-1;
  selection.deselectAll;
 end;
 if eType=edItems then begin
  hideAllActors;
  hideAllStones;
  showAllItems;
  ePlane:=-1;
 end;
 if eType=edActors then begin
  showAllStones;
  showAllItems;
  showAllActors;
  ePlane:=-1;
 end;
end;

procedure TO3Editor.tickEditorState;
begin
 if eType=edStones then begin
 end;
 if eType=edItems then begin
 end;
 if eType=edActors then begin
  if ePlane<>-1 then ePlane:=-1;
 end;
 if eType=edTest then begin
  updateEditorState;
 end;
end;

function TO3Editor.mouseToLevWindow;
begin
 Result:=true;
 mx:=mx+0.28;
 my:=my-0.2;
 if (mx>1)or(my>0.75)or(mx<-1)or(my<-0.75) then Result:=false;
end;

function TO3Editor.levWindowToLevPos;
begin
 Result:=true;
 mx:=-mx*8.65+cCenter[0];
 my:=my*8.65+cCenter[2];
 if (floor(mx)<0)or(floor(my)<0)or(floor(mx)>=owner.theLevel.FX)or(floor(my)>=owner.theLevel.FZ) then Result:=false;
end;

procedure TO3Editor.ExitEditor(sender:TO3Control);
begin
 doExit:=true;
end;

procedure TO3Editor.saveClick(sender:TO3Control);
var n,p:string;
begin
 if levselect.Execute('Save Level',p,n,curPack,curLev,menu) then
  if DirectoryExists(owner.path+LevelSubDir+p+PathDelim) then begin
   owner.theLevel.saveToFile(owner.path+LevelSubDir+p+PathDelim+n+'.o3','');
   curLev:=n;
   curPack:=p;
  end;
end;

procedure TO3Editor.stoneClick(sender:TO3Control);
begin
 eType:=edStones;
 updateEditorState;
end;

procedure TO3Editor.testClick(sender:TO3Control);
begin
 eType:=edTest;
end;

procedure TO3Editor.itemClick(sender:TO3Control);
begin
 eType:=edItems;
 updateEditorState;
end;

procedure TO3Editor.actorClick(sender:TO3Control);
begin
 eType:=edActors;
 updateEditorState;
end;

procedure TO3Editor.planeupClick(sender:TO3Control);
begin
 ePlane:=ePlane+1;
 if(ePlane>=owner.theLevel.FY) then
  ePlane:=-1;
end;

procedure TO3Editor.planedownClick(sender:TO3Control);
begin
 ePlane:=ePlane-1;
 if(ePlane<-1) then
  ePlane:=owner.theLevel.FY-1;
end;

procedure TO3Editor.copybufClick(sender:TO3Control);
begin
 selection.copyToBuffer;
end;

procedure TO3Editor.bufparamClick(sender:TO3Control);
begin
 if buffer<>nil then
  paramEdit(buffer);
end;

procedure TO3Editor.selectparamClick(sender:TO3Control);
var c:TGameObject;
begin
 if selection.fList.Count=1 then begin
  c:=TGameObject(selection.fList[0]);
  paramEdit(c);
  selection.fList.Clear;
  selection.select(c);
 end;
end;

procedure TO3Editor.linkClick(sender:TO3Control);
begin
 selection.deselectAll;
 linking:=true;
 linkSrc:=nil;
end;


procedure TO3Editor.RunEditor;
var sh:boolean;
    tick,delta:integer;
    lev:TLevel;
    mx,my:single;
    i,j:integer;
begin
 if owner.theLevel=nil then EXIT;
 menu.active:=true;
 owner.gfxSys.noswap:=true;
 owner.gfxSys.noclear:=true;
 sh:=owner.gfxSys.drawshadow;
 owner.gfxSys.DrawShadow:=false;
 tick:=owner.IOSys.getTicks;
 cPos[0]:=owner.theLevel.FX/2;
 cPos[1]:=owner.theLevel.FY+8;
 cPos[2]:=owner.theLevel.FZ/2;
 cCenter[0]:=owner.theLevel.FX/2;
 cCenter[1]:=0;
 cCenter[2]:=owner.theLevel.FZ/2;
 buffer:=o3objects.CreateStone('st-block',owner.theLevel);
 levelChanged:=false;
 while not (doExit or menu.exitMenu) do begin
  delta:=Integer(owner.IOSys.getTicks)-tick;
  tick:=owner.IOSys.getTicks;
  owner.InputSys.Tick;
  tickEditorState;
  if eType=edTest then begin
   owner.gfxSys.noswap:=false;
   owner.gfxSys.noclear:=false;
   owner.gfxSys.DrawShadow:=sh;
   menu.active:=false;

   lev:=owner.theLevel;
   lev.saveToFile(owner.path+EditorSubDir+'test.o3','');
   owner.StartLevelOxyd(owner.path+LevelSubDir+'import_ozone'+LibExt,owner.path+EditorSubDir+'test.o3','');
   owner.theLevel:=lev;

   menu.active:=true;
   owner.gfxSys.noswap:=true;
   owner.gfxSys.noclear:=true;
   sh:=owner.gfxSys.drawshadow;
   owner.gfxSys.DrawShadow:=false;
   eType:=edStones;
   updateEditorState;
  end;
  mx:=menu.mouseX;
  my:=menu.mouseY;
  if linking then begin
   if linkSrc=nil then menu.aToolTip:='Select the signal SOURCE (ESC: Cancel)!'
                  else menu.aToolTip:='Select the signal TARGET (ESC: Cancel)!'
  end
  else if (mouseToLevWindow(mx,my) and LevWindowToLevPos(mx,my)) then
  begin
   if (eType=edItems) or (eType=edStones) then begin
    if ePlane=-1 then menu.aToolTip:='X: '+intToStr(trunc(mx))+' Y: ALL'+' Z: '+intToStr(trunc(my))
    else menu.aToolTip:='X: '+intToStr(trunc(mx))+' Y: '+intToStr(ePlane)+' Z: '+intToStr(trunc(my));
   end
   else
    menu.aToolTip:='X: '+LeftStr(floatToStr(mx),5)+' Z: '+LeftStr(floatToStr(my),5)
  end;
  owner.gfxSys.clear;
  owner.gfxSys.viewport(-1.28,0.55,2,1.5);
  owner.gfxSys.setProjectionMatrix;
  owner.gfxSys.pushmatrix;
  owner.gfxSys.loadIdentityMatrix;
  owner.gfxSys.setPerspective(70,1.33,1,100);
  owner.gfxSys.setViewMatrix;
  owner.gfxSys.loadIdentityMatrix;
  owner.gfxSys.pushmatrix;
  if ePlane<>-1 then
   owner.gfxSys.lookAt(cPos[0],10,cPos[2],cCenter[0],cCenter[1],cCenter[2],0,1,1)
  else
   owner.gfxSys.lookAt(cPos[0],cPos[1],cPos[2],cCenter[0],cCenter[1],cCenter[2],0,1,1);
  owner.gfxSys.renderPass:=0;
  owner.gfxSys.translate(0,-abs(ePlane),0);
  owner.gfxSys.PaintLevel(owner.theLevel,delta,cPos[0],cPos[1],cPos[2],cCenter[0],cCenter[1],cCenter[2],ePlane,true);
  owner.gfxSys.disableLighting;
  owner.gfxSys.setColor(1,0,0,1);
  owner.gfxSys.disableDepthTest;
  if(owner.theLevel.signals.FList.Count>0) then begin
   for i:=0 to owner.theLevel.signals.FList.Count-1 do begin
    if(TSignal(owner.theLevel.signals.FList[i]).targets.Count>0) then
     for j:=0 to TSignal(owner.theLevel.signals.FList[i]).targets.Count-1 do
      with TSignal(owner.theLevel.signals.FList[i]) do
       owner.gfxSys.drawLine(source.cx,source.cy,source.cz,TGameObject(targets[j]).cx,TGameObject(targets[j]).cy,TGameObject(targets[j]).cz,3);
   end;
  end;
  owner.gfxSys.enableDepthTest;
  owner.gfxSys.enableLighting;
  owner.gfxSys.popmatrix;
  owner.gfxSys.setColor(1,1,1,1);
  if buffer<>nil then begin
   owner.gfxSys.viewport(0.65,-0.4,1.65,1.2);
   owner.gfxSys.pushmatrix;
   owner.gfxSys.lookAt(0,6,0,0,0,0,0,1,1);
   owner.gfxSys.setLightPos(0,6,0,1);
   owner.gfxSys.renderPass:=0;
   owner.gfxSys.translate(2,0,-2);
   if buffer.objectType=otActor then
    owner.gfxSys.translate(1.5,2,1.5);
   buffer.Render(delta);
//   owner.gfxSys.DrawModel(buffer.gfx,0,0,0,buffer.gfxType);
   owner.gfxSys.renderPass:=1;
   buffer.Render(0);
//   owner.gfxSys.DrawModel(buffer.gfx,0,0,0,buffer.gfxType);
   owner.gfxSys.popmatrix;
  end;
  owner.gfxSys.setProjectionMatrix;
  owner.gfxSys.popmatrix;
  owner.gfxSys.viewport(-1.33,1,2.66,2);
  owner.gfxSys.setViewMatrix;
  //if buffer<>nil then owner.gfxSys.DrawModel(buffer.gfx,1,-0.2,0,buffer.gfxType,nil,true);
//begin
  // menu.FindControl('exit').BgModel:=buffer.gfx;
  // menu.FindControl('exit').BgMType:=buffer.gfxType;
 // end;
  owner.gfxSys.DoMenu(menu);
  owner.gfxSys.DrawText(-1.33,-0.65,menu.aToolTip,0,0.75,1,1,1);
  owner.gfxSys.swap;
 end;
 if(levelChanged) then saveClick(nil);
 owner.gfxSys.noswap:=false;
 owner.gfxSys.noclear:=false;
 owner.gfxSys.DrawShadow:=sh;
end;

destructor TO3Editor.Destroy;
begin
 inherited Destroy;
 owner.InputSys.deleteListener(listener);
 listener.Free;
 selection.Free;
 levselect.Free;
 if buffer<>nil then buffer.free;
// if owner.theLevel<>nil then owner.theLevel.Free;
 menu.Free;
end;

constructor TO3Editor.Create;
var button:TO3CustomButton;
begin
 inherited Create;
 linking:=false;
 linkSrc:=nil;
 owner:=own;
 menu:=TO3Menu.Create(owner);
 menu.showToolTips:=false;

 button:=TO3CustomButton.Create(owner.gfxSys,'exit');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'button_exit.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=1.1;
 button.y:=-0.8;
 button.H:=0.3;
 button.W:=0.3;
 button.ToolTip:='Return to the Editor Menu';
 button.OnClick:=ExitEditor;
 menu.AddControl(button);


 button:=TO3CustomButton.Create(owner.gfxSys,'buffer');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.4,0.4,'buffer.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=1.05;
 button.y:=0.75;
 button.H:=0.3;
 button.W:=0.4;
 button.ToolTip:='Buffer Parameters (Hotkey: B)';
 button.OnClick:=bufparamClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'save');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'button_save.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=-1.2;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Save the Level (Hotkey: W)';
 button.OnClick:=saveClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'testlev');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'button_1pstart.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=-1.0;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Test the Level (Hotkey: T)';
 button.OnClick:=testClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'stoneed');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'stone_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=-0.8;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Edit Stones (Hotkey: S)';
 button.OnClick:=stoneClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'itemed');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'item_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=-0.6;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Edit Items (Hotkey: I)';
 button.OnClick:=itemClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'actored');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'actor_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=-0.4;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Edit Actors (Hotkey: A)';
 button.OnClick:=actorClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'paramb');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'param_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=-0.2;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Selection Parameters (Hotkey: P)';
 button.OnClick:=selectparamClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'attribb');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'attrib_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=0;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Level Attributes (Hotkey: V)';
 button.OnClick:=nil;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'planedown');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'planedown_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=0.2;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Edit lower Plane (Hotkey: PageDown)';
 button.OnClick:=planedownClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'planeup');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'planeup_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=0.4;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Edit upper Plane (Hotkey: PageUp)';
 button.OnClick:=planeupClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'copybuf');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'copybuf_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=0.6;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Copy Selection to Buffer (Hotkey: C)';
 button.OnClick:=copybufClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'link');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.2,0.2,'link_button.png',EditorSubDir);
 button.bgMtype:=plane;
 button.x:=0.8;
 button.y:=-0.8;
 button.H:=0.2;
 button.W:=0.2;
 button.ToolTip:='Create a Link (Hotkey: L)';
 button.OnClick:=LinkClick;
 menu.AddControl(button);


 processPaletteFiles;


 owner.gfxSys.FreePlaneXYModel(menu.mouseModel); // Free the standard mouse model
 menu.mouseModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'editor_mouse.png',BitmapSubDir); // load own mouse model


 doExit:=false;
 curPack:='test';
 curLev:='';
 levselect:=TO3EditorLevelSelector.Create(owner);

 listener:=TEditorListener.Create(self);
 owner.InputSys.addListener(listener);
 selection:=TEditorSelection.Create(self);
 menu.OnKeyDown:=listener.mkeyDown;
 menu.OnKeyUp:=listener.mkeyUp;
// if(FileExists(owner.path+LevelSubDir+curPack+PathDelim+curLev+'.o3')) then
//  owner.theLevel:=TLevel.Create(owner,owner.path+LevelSubDir+'import_ozone.so',owner.path+LevelSubDir+curPack+PathDelim+curLev+'.o3','nil',CreateStone,CreateItem,CreateActor)
// else begin
//  owner.theLevel:=nil;
///  EXIT;
// end;
 ePlane:=-1;
 eType:=edStones;
 updateEditorState;

end;

constructor TO3EditorMainMenu.create;
var button:TO3CustomButton;
begin
 self.owner:=owner;
 menu:=TO3Menu.Create(owner);
 menu.active:=false;

 button:=TO3CustomButton.Create(owner.gfxSys,'load');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.5,0.5,'button_save.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=-1;
 button.y:=0.7;
 button.H:=0.5;
 button.W:=0.5;
 button.ToolTip:='Load a Level';
 button.OnClick:=loadLevel;
 menu.AddControl(button);


 button:=TO3CustomButton.Create(owner.gfxSys,'new');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.5,0.5,'button_editor.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=1;
 button.y:=0.7;
 button.H:=0.5;
 button.W:=0.5;
 button.ToolTip:='Create a new Level';
 button.OnClick:=newLevel;
 menu.AddControl(button);


 button:=TO3CustomButton.Create(owner.gfxSys,'exit');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.5,0.5,'button_exit.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=1;
 button.y:=-0.7;
 button.H:=0.5;
 button.W:=0.5;
 button.ToolTip:='Exit Editor';
 button.OnClick:=exitMenu;
 menu.AddControl(button);
 
 menu.OnKeyDown:=keyDown;
end;

destructor TO3EditorMainMenu.destroy;
begin
 menu.Free;
end;

procedure TO3EditorMainMenu.run;
begin
 menu.active:=true;
 menu.exitMenu:=false;
 while not menu.exitMenu do begin
  owner.InputSys.Tick;
//  owner.gfxSys.clear;
  owner.gfxSys.DoMenu(menu);
//  owner.gfxSys.swap;
 end;
 menu.active:=false;
end;

procedure TO3EditorMainMenu.loadLevel(sender:TO3Control);
var levselect:TO3EditorLevelSelector;
    n,p:string;
begin
 levselect:=TO3EditorLevelSelector.Create(owner);
 menu.active:=false;
// if owner.theLevel<>nil then owner.theLevel.Free;
 if levselect.Execute('Load Level',p,n,'test','test1') and
    (FileExists(owner.path+LevelSubDir+p+PathDelim+n+'.o3')) then begin
  owner.theLevel:=TLevel.Create(owner,owner.path+LevelSubDir+'import_ozone'+LibExt,owner.path+LevelSubDir+p+PathDelim+n+'.o3','nil',CreateStone,CreateItem,CreateActor);
  owner.theLevel.editorMode:=true;
  editor:=TO3Editor.Create(owner);
  editor.curLev:=n;
  editor.curPack:=p;
  editor.RunEditor;
  editor.Free;
  if owner.theLevel<>nil then owner.theLevel.Free;
  owner.theLevel:=nil;
 end;
 menu.active:=true;
end;


procedure TO3EditorMainMenu.keyDown(sender:TO3Menu;key:TO3Key);
begin
 if key=O3K_ESCAPE then
  menu.exitMenu:=true;
end;


procedure TLevelSizeSelector.okClick(sender:TO3Control);
begin
 rlev:=1;
end;

procedure TLevelSizeSelector.cancelClick(sender:TO3Control);
begin
 rlev:=-1;
end;

function TLevelSizeSelector.Execute:boolean;
begin
 menu.active:=true;
 Result:=false;
 while rlev=0 do begin
  owner.InputSys.Tick;
  if owner.gfxSys.noclear then owner.gfxSys.clear;
  owner.gfxSys.setViewMatrix;
  owner.gfxSys.loadIdentityMatrix;
  owner.gfxSys.DoMenu(menu);
  if owner.gfxSys.noswap then owner.gfxSys.swap;
 end;
 if rlev>0 then Result:=true;
 Result:=Result and (x>0) and (y>0) and (z>0);
 menu.active:=false;
end;

procedure TLevelSizeSelector.ActionX(sender:TO3Control);
begin
 x:=StrToInt(TO3MouseTextControl(sender).Text);
end;

procedure TLevelSizeSelector.ActionY(sender:TO3Control);
begin
 y:=StrToInt(TO3MouseTextControl(sender).Text);
end;

procedure TLevelSizeSelector.ActionZ(sender:TO3Control);
begin
 z:=StrToInt(TO3MouseTextControl(sender).Text);
end;

constructor TLevelSizeSelector.Create(own:TOzoneMainClass);
var button:TO3CustomButton;
    inp:TO3UIntEdit;
begin
 inherited Create;
 owner:=own;
 rlev:=0;
 x:=5;
 y:=2;
 z:=5;
 menu:=TO3Menu.Create(owner);

 button:=TO3CustomButton.Create(owner.gfxSys,'iXd');
 button.x:=0;
 button.y:=0.8;
 button.H:=0.1;
 button.W:=2;
 button.Text:='X Size';
 button.TextSize:=0.8;
 menu.AddControl(button);

 inp:=TO3UIntEdit.Create(owner.gfxSys,'inpX');
 inp.x:=0;
 inp.y:=0.7;
 inp.Text:=IntToStr(x);
 inp.OnAction:=ActionX;
 menu.AddControl(inp);


 button:=TO3CustomButton.Create(owner.gfxSys,'iYd');
 button.x:=0;
 button.y:=0.4;
 button.H:=0.1;
 button.W:=2;
 button.Text:='Y Size';
 button.TextSize:=0.8;
 menu.AddControl(button);

 inp:=TO3UIntEdit.Create(owner.gfxSys,'inpY');
 inp.x:=0;
 inp.y:=0.3;
 inp.Text:=IntToStr(y);
 inp.OnAction:=ActionY;
 menu.AddControl(inp);


 button:=TO3CustomButton.Create(owner.gfxSys,'iZd');
 button.x:=0;
 button.y:=0;
 button.H:=0.1;
 button.W:=2;
 button.Text:='Z Size';
 button.TextSize:=0.8;
 menu.AddControl(button);

 inp:=TO3UIntEdit.Create(owner.gfxSys,'inpZ');
 inp.x:=0;
 inp.y:=-0.1;
 inp.Text:=IntToStr(z);
 inp.OnAction:=ActionZ;
 menu.AddControl(inp);


 button:=TO3CustomButton.Create(owner.gfxSys,'okB');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.4,0.4,'button_ok.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=0.8;
 button.y:=-0.6;
 button.H:=0.4;
 button.W:=0.4;
 button.OnClick:=OkClick;
 menu.AddControl(button);

 button:=TO3CustomButton.Create(owner.gfxSys,'cancelB');
 button.bgModel:=owner.gfxSys.CreatePlaneXYModel(0.4,0.4,'button_cancel.png',BitmapSubDir);
 button.bgMtype:=plane;
 button.x:=-0.8;
 button.y:=-0.7;
 button.H:=0.4;
 button.W:=0.4;
 button.OnClick:=CancelClick;
 menu.AddControl(button);

 owner.gfxSys.FreePlaneXYModel(menu.mouseModel); // Free the standard mouse model
 menu.mouseModel:=owner.gfxSys.CreatePlaneXYModel(0.3,0.3,'editor_mouse.png',BitmapSubDir); // load own mouse model

 menu.showToolTips:=false;
 menu.active:=false;
end;

destructor TLevelSizeSelector.Destroy;
begin
 menu.Free;
 inherited Destroy;
end;




procedure TO3EditorMainMenu.newLevel(sender:TO3Control);
var x,y,z:integer;
    sizer:TLevelSizeSelector;
begin
 menu.active:=false;
 sizer:=TLevelSizeSelector.Create(owner);
 if not sizer.Execute then begin
  sizer.Free;
  menu.active:=true;
  EXIT;
 end;
 menu.active:=true;
 owner.theLevel:=TLevel.Create(owner,'nil','','',CreateStone,CreateItem,CreateActor);
 owner.theLevel.editorMode:=true;
 owner.theLevel.FX:=sizer.x;
 owner.theLevel.FY:=sizer.y;
 owner.theLevel.FZ:=sizer.z;
 setLength(owner.theLevel.FCells,owner.theLevel.FX,owner.theLevel.FY,owner.theLevel.FZ);
 for x:=0 to owner.theLevel.FX-1 do
  for y:=0 to owner.theLevel.FY-1 do
   for z:=0 to owner.theLevel.FZ-1 do
    owner.theLevel.FCells[x,y,z]:=TCell.Create('nil','nil',owner.theLevel);
 sizer.Free;
 editor:=TO3Editor.Create(owner);
 editor.curLev:='unnamed';
 editor.curPack:='test';
 menu.active:=false;
 owner.theLevel.loadingLib:=TLevelLib.Create(owner.path+LevelSubDir+'import_ozone'+LibExt);
 editor.RunEditor;
 editor.Free;
 if owner.theLevel<>nil then owner.theLevel.Free;
 owner.theLevel:=nil;
 menu.active:=true;
end;

procedure TO3EditorMainMenu.exitMenu(sender:TO3Control);
begin
 menu.exitMenu:=true;
end;

end.

