unit o3menu;

interface
 uses SysUtils,o3interfaces,o3utils,Classes;
type
 TO3Control = class;
 TNotify = procedure(Sender:TO3Control) of object;
 TO3Control = class
  private
   FX,FY,FW,FH:Single;
   gfxSys:TO3GfxSystem;
   FName:string;
   procedure SetX(const val:single); virtual;
   procedure SetY(const val:single); virtual;
   function GetX:Single; virtual;
   function GetY:Single; virtual;
   procedure SetW(const val:single); virtual;
   procedure SetH(const val:single); virtual;
   function GetW:Single; virtual;
   function GetH:Single; virtual;
   procedure LoadBGModel; virtual;
   procedure FreeBGModel; virtual;
   procedure OnCreate; virtual;
  public
   OnAction:TNotify;
   Tooltip:string;
   Owner:Pointer;
   BgModel:pointer;
   BgMType:TModelType;
   BgFPS:Single;
   BgLoop:boolean;
   property X:Single read GetX write SetX;
   property Y:Single read GetY write SetY;
   property W:Single read GetW write SetW;
   property H:Single read GetH write SetH;
   property name:string read FName;
   constructor Create(gfx:TO3GfxSystem;name:string); virtual;
   destructor Destroy; override;
   procedure Render(dtime:integer); virtual;
   function isInBounds(x,y:single):Boolean; virtual;
 end;
 TO3TextControl = class(TO3Control)
  private
   FTextSize,FFontR,FFontG,FFontB,FTextX,FTextY:Single;
   FText:string;
   FFontSet:Cardinal;
   procedure setText(const F:string); virtual;
   function getText:string; virtual;
  public
   WordWarp:boolean;
   property Text:string read getText write setText;
   property TextSize:single read FTextSize write FTextSize;
   procedure Render(dTime:integer); override;
 end;
 TO3MouseTextControl = class(TO3TextControl)
  public
   OnClick:TNotify;
   procedure Click(x,y:Single); virtual;
 end;
 TO3MouseTextKeyControl = class(TO3MouseTextControl)
  public
   procedure KeyDown(key:TO3Key); virtual;
   procedure KeyUp(key:TO3Key); virtual;
 end;
 TO3TextEdit = class(TO3MouseTextKeyControl)
  private
   procedure LoadBgModel; override;
   procedure FreeBgModel; override;
  public
   procedure KeyDown(key:TO3Key); override;
   procedure KeyUp(key:TO3Key); override;
 end;
 TO3RichEdit = class(TO3TextEdit)
  private
   shift:boolean;
   caps:boolean;
   procedure LoadBgModel; override;
   procedure FreeBgModel; override;
  public
   procedure KeyDown(key:TO3Key); override;
   procedure KeyUp(key:TO3Key); override;
 end;
 TO3UIntEdit = class(TO3TextEdit)
  private
   procedure LoadBgModel; override;
   function isValid:boolean; virtual;
   procedure setText(const F:string); override;
  public
   procedure KeyDown(key:TO3Key); override;
 end;
 TO3IntEdit = class(TO3UIntEdit)
  private
   procedure LoadBgModel; override;
   function isValid:boolean; override;
  public
   procedure KeyDown(key:TO3Key); override;
 end;
 TO3FloatEdit = class(TO3IntEdit)
  private
   procedure LoadBgModel; override;
   function isValid:boolean; override;
  public
   procedure KeyDown(key:TO3Key); override;
 end;
 TO3CheckBox = class(TO3MouseTextControl)
  private
   bgmodel_t:Pointer;
   bgmodel_f:Pointer;
   procedure LoadBgModel; override;
   procedure FreeBgModel; override;
  public
   value:boolean;
   procedure Render(dTime:integer); override;
   procedure Click(x,y:single); override;
 end;

 TO3CustomButton = class(TO3MouseTextControl)
  private
   procedure OnCreate; override;
 end;
 TO3Button = class(TO3CustomButton)
  private
   procedure LoadBgModel; override;
   procedure FreeBgModel; override;
 end;
 TO3Container = class
  private
   FControls:TList;
   ozone:Pointer;
  public
   procedure OnCreate; virtual;
   procedure OnDestroy; virtual;
   procedure AddControl(c:TO3Control; own:Boolean = true); virtual;
   procedure DeleteControl(c:TO3Control); virtual;
   function FindControl(name:string):TO3Control; virtual;
   constructor Create(o3:pointer); virtual;
   destructor Destroy; override;
   procedure Render(dTime:integer); virtual;
 end;
 TO3Menu = class;
 TO3MenuListener = class(TO3InputListener)
  owner:TO3Menu;
  procedure keyDown(key:TO3Key); override;
  procedure keyUp(key:TO3Key); override;
  procedure mouseDown(button:integer); override;
  procedure mouseMove(dx,dy:integer); override;
 end;
 TO3KeyProc = procedure(sender:TO3Menu; key:TO3Key) of object;
 TO3Menu = class(TO3Container)
 private
  mx,my:single;
  mmulti:single;
  listener:TO3MenuListener;
 public
  aToolTip:string;
  exitMenu:boolean;
  showToolTips:boolean;
  active:boolean;
  MouseModel:Pointer;
  OnKeyDown:TO3KeyProc;
  OnKeyUp:TO3KeyProc;
  procedure OnCreate; override;
  procedure OnDestroy; override;
  procedure keyDown(key:TO3Key);
  procedure keyUp(key:TO3Key);
  procedure mouseDown(button:integer);
  procedure mouseMove(dx,dy:integer);
  procedure Render(dTime:integer); override;
  property mouseX:single read mx;
  property mouseY:single read my;
 end;
implementation
 uses md2,o3m,ac3d,o3game;
 {*************** TO3MenuListener *****************}
 procedure TO3MenuListener.keyDown;
 begin
  if owner.active then owner.keyDown(key);
 end;
 procedure TO3MenuListener.keyUp;
 begin
  if owner.active then owner.keyUp(key);
 end;
 procedure TO3MenuListener.mouseDown;
 begin
  if owner.active then owner.mouseDown(button);
 end;
 procedure TO3MenuListener.mouseMove;
 begin
  if owner.active then owner.mouseMove(dx,dy);
 end;
 {******************* TO3Menu *********************}
 procedure TO3Menu.OnCreate;
 begin
  showToolTips:=true;
  exitMenu:=false;
  mx:=0;
  my:=0;
  mmulti:=1/300;
  mouseModel:=TOzoneMainClass(ozone).gfxSys.CreatePlaneXYModel(0.1,0.1,'menu_mouse.png',BitmapSubDir);//TO3MFile.Create(ozone);
//  TO3MFile(mouseModel).LoadFromFile(TOzoneMainClass(ozone).path+MenuModelSubDir+'mouse.o3m');
  listener:=TO3MenuListener.Create;
  listener.owner:=self;
  TOzoneMainClass(ozone).InputSys.addListener(listener);
 end;
 procedure TO3Menu.OnDestroy;
 begin
  TOzoneMainClass(ozone).InputSys.deleteListener(listener);
  listener.Free;
  //TO3MFile(mouseModel).Free;
  TOzoneMainClass(ozone).gfxSys.FreePlaneXYModel(mouseModel);
 end;
 procedure TO3Menu.Render;
 begin
  inherited;
  TOzoneMainClass(ozone).gfxSys.DrawModel(mouseModel,mx,my,0,plane);
 end;
 procedure TO3Menu.keyDown;
 var i:integer;
     b:boolean;
 begin
  b:=true;
  if FControls.Count<=0 then EXIT;
  for i:=0 to FControls.Count-1 do begin
   if FControls[i]=nil then CONTINUE;
   if TO3Control(FControls[i]) is TO3MouseTextKeyControl then
    if TO3MouseTextKeyControl(FControls[i]).isInBounds(mx,my) then begin
     TO3MouseTextKeyControl(FControls[i]).KeyDown(key);
     b:=false;
    end;
  end;
  if Assigned(OnKeyDown) and b then
   onKeyDown(self,key);
 end;
 procedure TO3Menu.keyUp;
 var i:integer;
     b:boolean;
 begin
  b:=true;
  if FControls.Count<=0 then EXIT;
  for i:=0 to FControls.Count-1 do begin
   if FControls[i]=nil then CONTINUE;
   if TO3Control(FControls[i]) is TO3MouseTextKeyControl then
    if TO3MouseTextKeyControl(FControls[i]).isInBounds(mx,my) then begin
     TO3MouseTextKeyControl(FControls[i]).KeyUp(key);
     b:=false;
    end;
  end;
  if Assigned(OnKeyUp) and b then
   onKeyUp(self,key);
 end;
 procedure TO3Menu.mouseDown;
 var i:integer;
 begin
  if FControls.Count<=0 then EXIT;
  for i:=0 to FControls.Count-1 do begin
   if FControls[i]=nil then CONTINUE;
   if TO3Control(FControls[i]) is TO3MouseTextControl then
    if TO3MouseTextControl(FControls[i]).isInBounds(mx,my) then
     TO3MouseTextControl(FControls[i]).Click(mx,my);
  end;
 end;
 procedure TO3Menu.mouseMove;
 var i:integer;
 begin
  mx:=mx+dx*mmulti;
  my:=my-dy*mmulti;
  if (mx<-1.33) then mx:=-1.33;
  if (mx>1.33) then mx:=1.33;
  if (my<-1) then my:=-1;
  if (my>1) then my:=1;
  aToolTip:='';
  if FControls.Count<=0 then EXIT;
  for i:=0 to FControls.Count-1 do begin
   if FControls[i]=nil then CONTINUE;
   if TO3Control(FControls[i]) is TO3MouseTextControl then
    if TO3MouseTextControl(FControls[i]).isInBounds(mx,my) then
     aToolTip:=TO3MouseTextControl(FControls[i]).Tooltip;
  end;
 end;
 {***************** TO3Container ******************}
 procedure TO3Container.OnCreate;
 begin
 end;
 procedure TO3Container.OnDestroy;
 begin
 end;
 procedure TO3Container.AddControl;
 begin
  FControls.Add(c);
  if own then
   c.Owner:=Self;
 end;
 function TO3Container.FindControl;
 var x:integer;
 begin
  Result:=nil;
  if FControls.Count>0 then
   for x:=0 to FControls.Count-1 do
    if (FControls[x]<>nil)and(TO3Control(FControls[x]).FName=name) then begin
     Result:=FControls[x];
     EXIT;
    end;
 end;
 procedure TO3Container.DeleteControl;
 begin
  FControls.Remove(c);
  if c.Owner=self then
   c.Free;
 end;
 constructor TO3Container.Create;
 begin
  inherited Create();
  ozone:=o3;
  FControls:=TList.Create;
  OnCreate;
 end;
 destructor TO3Container.Destroy;
 var x:integer;
 begin
  if FControls.Count>0 then
   for x:=0 to FControls.Count-1 do
    if (FControls[x]<>nil)and(TO3Control(FControls[x]).Owner=Self) then
     TO3Control(FControls[x]).Free;
  FControls.Clear;
  FControls.Free;
  OnDestroy;
  inherited;
 end;
 procedure TO3Container.Render;
 var x:integer;
 begin
  if FControls.Count<=0 then Exit;
  for x:=0 to FControls.Count-1 do begin
   TO3Control(FControls[x]).Render(dTime);
  end;
 end;
 {****************** TO3Control *******************}
 procedure TO3Control.OnCreate;
 begin
 end;
 function TO3Control.isInBounds;
 begin
  Result:=false;
  if (x>FX-FW/2)and(x<FX+FW/2)and(y>FY-FH/2)and(y<FY+FH/2) then Result:=true;
 end;
 procedure TO3Control.Render;
 begin
  if bgModel=nil then exit;
  if bgMType=quake then
   TQuakeModel(bgModel).animationStep(dtime/1000,bgFPS,bgLoop);
  gfxSys.DrawModel(bgModel,X,Y,0,bgMType);
 end;
 procedure TO3Control.LoadBGModel;
 begin
 end;
 procedure TO3Control.FreeBGModel;
 begin
  if bgModel=nil then exit;
  case bgMtype of
   ac: TAC3dFile(bgModel).Free;
   quake: TQuakeModel(bgModel).Free;
   plane: gfxSys.FreePlaneXYModel(bgModel);
   o3utils.ozone: TO3MFile(bgModel).Free;
  end;
 end;
 constructor TO3Control.Create;
 begin
  inherited Create;
  bgModel:=nil;
  onAction:=nil;
  FName:=Name;
  FX:=0;
  FY:=0;
  FH:=0.6;
  FW:=0.6;
  gfxSys:=gfx;
  OnCreate;
  LoadBGModel;
 end;
 destructor TO3Control.Destroy;
 begin
  FreeBGModel;
  inherited;
 end;
 procedure TO3Control.SetX;
 begin
  FX:=val;
 end;
 procedure TO3Control.SetY;
 begin
  FY:=val;
 end;
 function TO3Control.GetX;
 begin
  Result:=FX;
 end;
 function TO3Control.GetY;
 begin
  Result:=FY;
 end;
 procedure TO3Control.SetW;
 begin
  FW:=val;
 end;
 procedure TO3Control.SetH;
 begin
  FH:=val;
 end;
 function TO3Control.GetW;
 begin
  Result:=FW;
 end;
 function TO3Control.GetH;
 begin
  Result:=FX;
 end;
 {********************* TO3TextControl *********************}
 procedure TO3TextControl.setText;
 begin
  FText:=f;
 end;
 function TO3TextControl.getText;
 begin
  Result:=FText;
 end;
 procedure TO3TextControl.Render;
 var w,l:string;
     i:integer;
     yt:single;
 begin
  inherited Render(dTime);
//  gfxsys.pushmatrix;
  if FText='' then exit;
  if not wordWarp then gfxSys.DrawText(FX+FTextX-Length(FText)*8*0.005*FTextSize,FY+FTextY,FText,FFontSet,FTextSize,FFontR,FFontG,FFontB,false)
  else begin
   i:=1;
   l:='';
   w:=GetToken(FText,i,' ');
   yt:=FTextY;
   while w<>'' do begin
    if (Length(l+w+' ')*8*0.01*FTextSize)>=FW then begin
     gfxSys.DrawText(FX+FTextX-Length(l)*8*0.005*FTextSize,FY+yt,l,FFontSet,FTextSize,FFontR,FFontG,FFontB,false);
     yt:=yt-8*0.01*FTextSize;
     l:='';
    end;
    l:=l+w+' ';
    inc(i);
    w:=GetToken(FText,i,' ');
    if w='' then
     gfxSys.DrawText(FX+FTextX-Length(l)*8*0.005*FTextSize,FY+yt,l,FFontSet,FTextSize,FFontR,FFontG,FFontB,false);
   end;
  end;
//  gfxsys.popmatrix;
 end;
 {****************** TO3TextMouseControl *******************}
 procedure TO3MouseTextControl.Click;
 begin
  if Assigned(OnClick) then
   OnClick(Self);
  if Assigned(OnAction) then
   OnAction(Self);
 end;
 {******************** TO3CustomButton *********************}
 procedure TO3CustomButton.OnCreate;
 begin
  WordWarp:=false;
  FTextSize:=1;
  FFontR:=1;
  FFontG:=1;
  FFontB:=1;
  FTextX:=0.0;
  FTextY:=0.0;
  FFontSet:=0;
 end;
 {*********************** TO3Button ************************}
 procedure TO3Button.LoadBgModel;
 begin
  writeln('1');
//  bgMType:=ac;
//  bgModel:=TAC3DFile.Create(gfxSys.owner);
//  TAC3DFile(bgModel).LoadFromFile(TOzoneMainClass(gfxSys.owner).path+MenuModelSubDir+'button.ac');
  writeln('2');
 end;
 procedure TO3Button.FreeBgModel;
 begin
//  TAC3DFile(bgModel).Free;
 end;

 {*************** TO3MouseTextKeyControl ********************}
 procedure TO3MouseTextKeyControl.keyDown;
 begin
 end;
 procedure TO3MouseTextKeyControl.keyUp;
 begin
 end;

 {************************ TO3RichEdit **********************}

  procedure TO3RichEdit.LoadBgModel;
  begin
   shift:=false;
   caps:=false;
   bgMType:=plane;
   bgModel:=gfxSys.CreatePlaneXYModel(2,0.4,'textinput.png',BitmapSubDir);
   H:=0.4;
   W:=2;
   FFontR:=1;
   FFontG:=1;
   FFontB:=1;
   FTextX:=0;
   FTextY:=0.08;
   FTextSize:=0.75;
   FFontSet:=0;
   WordWarp:=true;
  end;

  procedure TO3RichEdit.FreeBgModel;
  begin
   gfxSys.FreePlaneXYModel(bgModel);
  end;

  procedure TO3RichEdit.KeyDown(key:TO3Key);
  begin
   if (key=O3K_LSHIFT)or(key=O3K_RSHIFT) then
    shift:=true;
   if (key=O3K_CAPSLOCK) then
    caps:=not caps;
   if(ord(key)>31)and(ord(key)<=127) then begin
    if (shift xor caps) then FText:=FText+UpperCase(Chr(Ord(key)))
    else FText:=FText+Chr(Ord(key));
   end;
   if key=O3K_BACKSPACE then Text:=Copy(Text,1,Length(Text)-1);
   if Assigned(OnAction) then
    OnAction(Self);
  end;

  procedure TO3RichEdit.KeyUp(key:TO3Key);
  begin
   if (key=O3K_LSHIFT)or(key=O3K_RSHIFT) then
    shift:=false;
  end;

 {************************ TO3TextEdit **********************}
  procedure TO3TextEdit.LoadBgModel;
  begin
   bgMType:=plane;
   bgModel:=gfxSys.CreatePlaneXYModel(2,0.2,'textinput.png',BitmapSubDir);
   H:=0.2;
   W:=2;
   FFontR:=1;
   FFontG:=1;
   FFontB:=1;
   FTextX:=0;
   FTextY:=-0.03;
   FTextSize:=1;
   FFontSet:=0;
 end;
  
  procedure TO3TextEdit.FreeBgModel;
  begin
   gfxSys.FreePlaneXYModel(bgModel);
  end;

  //procedure TO3TextEdit.Render(dTime:integer);
  //begin
//   inherited Render(dTime);
  //end;

  procedure TO3TextEdit.KeyDown(key:TO3Key);
  begin
//   writeln('TextEdit key down: char '+Chr(Ord(key))+' ord '+inttostr(ord(key)));
   if(ord(key)>32)and(ord(key)<=127) then
    FText:=FText+Chr(Ord(key));
   if key=O3K_BACKSPACE then Text:=Copy(Text,1,Length(Text)-1);
   if Assigned(OnAction) then
    OnAction(Self);
  end;

  procedure TO3TextEdit.KeyUp(key:TO3Key);
  begin
  end;
// ------------------------ TO3UIntEdit -------------------------
 procedure TO3UIntEdit.setText;
 begin
  FText:=Copy(F,1,5);
 end;

 function TO3UIntEdit.isValid:boolean;
 begin
  Result:=false;
  if Length(FText)>0 then Result:=true;
 end;
 
 procedure TO3UIntEdit.LoadBgModel;
 begin
  inherited;
  H:=0.15;
  W:=1;
  gfxSys.FreePlaneXYModel(bgModel);
  bgMType:=plane;
  bgModel:=gfxSys.CreatePlaneXYModel(1,0.15,'uintinput.png',BitmapSubDir);
 end;

 procedure TO3UIntEdit.KeyDown(key:TO3Key);
 begin
  if key=O3K_BACKSPACE then Text:=Copy(Text,1,Length(Text)-1);
  if Length(FText)>6 then exit;
  if((ord(key)>=Ord(O3K_0))and(ord(key)<=ord(O3K_9))) or
    ((ord(key)>=Ord(O3K_KP0))and(ord(key)<=ord(O3K_KP9))) then
   FText:=FText+Chr(Ord(key));
  if isValid and Assigned(OnAction) then OnAction(Self);
 end;


// ------------------------- TO3IntEdit -------------------------
 function TO3IntEdit.isValid:boolean;
 begin
  Result:=inherited isValid;
  if Result and (Length(FText)=1) and (FText[1]='-') then Result:=false;
 end;

 procedure TO3IntEdit.LoadBgModel;
 begin
  inherited;
  gfxSys.FreePlaneXYModel(bgModel);
  bgMType:=plane;
  bgModel:=gfxSys.CreatePlaneXYModel(1,0.15,'intinput.png',BitmapSubDir);
 end;

 procedure TO3IntEdit.KeyDown(key:TO3Key);
 begin
  inherited;
  if((key=O3K_MINUS)or(key=O3K_KP_MINUS))and(Length(FText)=0) then
   FText:=FText+'-';
 end;
// ------------------------ TO3FloatEdit ------------------------
 function TO3FloatEdit.isValid:boolean;
 begin
  Result:=inherited isValid;
  if Result and (FText[Length(FText)]='.') then Result:=false;
 end;

 procedure TO3FloatEdit.LoadBgModel;
 begin
  inherited;
  gfxSys.FreePlaneXYModel(bgModel);
  bgMType:=plane;
  bgModel:=gfxSys.CreatePlaneXYModel(1,0.15,'floatinput.png',BitmapSubDir);
 end;

 procedure TO3FloatEdit.KeyDown(key:TO3Key);
 begin
  inherited;
  if((key=O3K_PERIOD)or(key=O3K_KP_PERIOD))and(Pos(FText,'.')=0) then begin
   if Length(FText)>0 then FText:=FText+'.'
   else FText:='0.';
  end;
 end;

{ TO3CheckBox = class(TO3MouseTextControl)
  private
   procedure LoadBgModel; override;
   procedure FreeBgModel; override;
  public
   procedure Click(x,y:single); override;
 end;}
 procedure TO3CheckBox.LoadBgModel;
 begin
  bgModel:=nil;
  bgModel_f:=gfxSys.CreatePlaneXYModel(0.09,0.09,'checkbox_false.png',BitmapSubDir);
  bgModel_t:=gfxSys.CreatePlaneXYModel(0.09,0.09,'checkbox_true.png',BitmapSubDir);
  FFontR:=1;
  FFontG:=1;
  FFontB:=1;
  FTextSize:=0.8;
  FTextY:=-0.03;
  FFontSet:=0;
 end;

 procedure TO3CheckBox.FreeBgModel;
 begin
  gfxSys.FreePlaneXYModel(bgModel_f);
  gfxSys.FreePlaneXYModel(bgModel_t);
 end;

 procedure TO3CheckBox.Render;
 begin
  inherited;
  if value then gfxSys.DrawModel(bgModel_t,FX+FTextX-Length(FText)*8*0.005*FTextSize-0.12,FY,0,plane)
  else gfxSys.DrawModel(bgModel_f,FX+FTextX-Length(FText)*8*0.005*FTextSize-0.12,FY,0,plane);
 end;

 procedure TO3CheckBox.Click(x,y:single);
 begin
  value:=not value;
  inherited;
 end;


end.
